# -*- coding: utf-8 -*-

"""
Zenoss device_router
"""

import re
import warnings
from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter
from zenossapi.routers.properties import PropertiesRouter
from zenossapi.routers.template import TemplateRouter


class DeviceRouter(ZenossRouter):
    """
    Class for interacting with the Zenoss device router
    """

    def __init__(self, url, headers, ssl_verify):
        super(DeviceRouter, self).__init__(url, headers, ssl_verify,
                                           'device_router', 'DeviceRouter')
        self.uid = None
        self.properties = None
        self.prod_states_by_value = {}
        self.prod_states_by_name = {}
        self.priorities_by_value = {}
        self.priorities_by_name = {}

        prod_states_data = self._router_request(
            self._make_request_data(
                'getProductionStates',
                dict(),
            )
        )
        for state in prod_states_data['data']:
            self.prod_states_by_name[state['name']] = state['value']
            self.prod_states_by_value[state['value']] = state['name']

        priority_data = self._router_request(
            self._make_request_data(
                'getPriorities',
                dict(),
            )
        )
        for p in priority_data['data']:
            self.priorities_by_name[p['name']] = p['value']
            self.priorities_by_value[p['value']] = p['name']

    def __repr__(self):
        if self.uid:
            identifier = self.uid
        else:
            identifier = hex(id(self))

        return '<{0} object at {1}>'.format(
            type(self).__name__, identifier
        )

    def _find_nodes_in_tree(self, tree_data):
        """
        Works through the dict structure returned by the Zenoss API for
        a device class tree and returns the nodes.

        Arguments:
            tree_data (dict): Templates data returned by the API

        Returns:
            list:
        """
        tree = []
        for node in tree_data['children']:
            node_children = self._find_nodes_in_tree(node)
            tree.append(
                dict(
                    uid=node['path'],
                    children=node_children
                )
            )

        return tree

    def _get_info_by_uid(self, uid):
        """
        Get object properties by the full UID

        Arguments:
            uid (str): UID for the object

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'getInfo',
                dict(uid=uid),
            )
        )

    def _get_device_by_uid(self, device_uid):
        """
        Get a device by its full UID

        Arguments:
            device_uid (str): The uid of the device to get

        Returns:
            ZenossDevice:
        """
        device_data = self._get_info_by_uid(device_uid)

        return ZenossDevice(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            device_data['data']
        )

    def _move_devices_by_uid(self, devices, device_class):
        """
        Move the devices in the list to a new device class.

        Arguments:
            devices (list): List of device uids to move
            device_class (str): Target device class for the move

        Returns:
            list(dict(str, str)): List of Job manager info for each device move ::

            [{
                'uuid': Job manager uuid for the device move,
                'description': Description of the move job,
            }]

        """
        if not isinstance(devices, list):
            raise ZenossAPIClientError(
                'Type error: devices to move must be a list')

        move_job_data = self._router_request(
            self._make_request_data(
                'moveDevices',
                dict(
                    uids=devices,
                    target=device_class,
                ),
            )
        )
        move_jobs = []
        for mj in move_job_data['new_jobs']:
            move_jobs.append(dict(
                uuid=mj['uuid'],
                description=mj['description']
            ))

        return move_jobs

    def _add_device_class(self, new_class, parent,
                          description='', connection_info=None):
        """
        Add a new device class under a parent path

        Arguments:
            new_class (str): Name of the device class to add
            parent (str): Device class to place the new class under
            description (str): Description for the new class
            connection_info (list): zProperties that are the credentials for
                accessing this device class, e.g. zCommandUsername,
                zCommandPassword

        Returns:
            ZenossDeviceClass:
        """
        response_data = self._router_request(
            self._make_request_data(
                'addDeviceClassNode',
                dict(
                    type='organizer',
                    contextUid=parent,
                    id=new_class,
                    description=description,
                    connectionInfo=connection_info,
                )
            )
        )

        dc_data = self._get_info_by_uid(response_data['nodeConfig']['path'])
        return ZenossDeviceClass(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            dc_data['data']
        )

    def _lock_devices_by_uid(self, devices, updates=False, deletion=False,
                             send_event=False):
        """
        Lock devices from changes.

        Arguments:
            devices (list): Uids of the devices to lock
            updates (bool): Lock devices from updates
            deletion (bool): Lock devices from deletion
            send_event (bool): Send an event when an action is blocked by locking

        Returns:
            str: Response message
        """
        response_data = self._router_request(
            self._make_request_data(
                'lockDevices',
                dict(
                    uids=devices,
                    hashcheck='',
                    updates=updates,
                    deletion=deletion,
                    sendEvent=send_event,
                )
            )
        )

        return response_data['msg']

    def _reset_device_ip_addresses_by_uid(self, devices, ip_address=''):
        """
        Reset IP addresses of devices, either by DNS lookup or manually
        specified address

        Arguments:
            devices (list): Uids of devices to reset IP address for
            ip_address (str): IP address to set device to
        """
        response_data = self._router_request(
            self._make_request_data(
                'resetIp',
                dict(
                    uids=devices,
                    hashcheck='',
                    ip=ip_address,
                )
            )
        )

        return response_data['msg']

    def _set_production_state_by_uids(self, devices, production_state):
        """
        Sets the production state of a list of devices.

        Arguments:
            devices (list): Uids of devices to set production state for
            production_state (int): Numeric value of production state to set

        Returns:
            str: Response message
        """
        if production_state not in self.prod_states_by_value:
            raise ZenossAPIClientError(
                'Production state {0} is not a valid option'.format(
                    production_state))

        response_data = self._router_request(
            self._make_request_data(
                'setProductionState',
                dict(
                    uids=devices,
                    hashcheck='',
                    prodState=production_state,
                )
            )
        )

        return response_data['msg']

    def _set_priority_by_uids(self, devices, priority):
        """
        Sets the priority of a list of devices

        Arguments:
            devices (list): Uids of devices to set priority for
            priority (int): Numeric value of priority to set

        Returns:
            str: Response message
        """
        if priority not in self.priorities_by_value:
            raise ZenossAPIClientError(
                "Priority {0} is not a valid option".format(
                    priority
                )
            )

        response_data = self._router_request(
            self._make_request_data(
                'setPriority',
                dict(
                    uids=devices,
                    hashcheck='',
                    priority=priority,
                )
            )
        )

        return response_data['msg']

    def _set_collector_by_uids(self, devices, collector):
        """
        Sets the collector for a list of devices.

        Arguments:
            devices (list): Uids of the devices to set collector for
            collector (str): The collector to set devices to use

        Returns:
            list(dict(str, str)): List of Job manager info for each device move ::

            [{
                'uuid': Job manager uuid for the device move
                'description': Description of the move job
            }]

        """
        collector_job_data = self._router_request(
            self._make_request_data(
                'setCollector',
                dict(
                    uids=devices,
                    hashcheck='',
                    collector=collector,
                )
            )
        )

        return dict(
            uuid=collector_job_data['new_jobs']['uuid'],
            description=collector_job_data['new_jobs']['description']
        )

    def _delete_devices_by_uid(self, devices, action, del_events=False,
                               del_perf=False):
        """
        Remove a list of devices from their organizer UID, or delete them
        from Zenoss altogether.

        Arguments:
            devices (list): Uids of devices to remove/delete
            action (str): 'remove' to remove the devices from their organizer,
                'delete' to delete them from Zenoss
            del_events (bool): Remove all events for the devices
            del_perf (bool): Remove all perf data for the devices

        Returns:
            dict:
        """
        if action not in ['remove', 'delete']:
            raise ZenossAPIClientError(
                "Delete action must be either 'remove' or 'delete'"
            )

        response_data = self._router_request(
            self._make_request_data(
                'removeDevices',
                dict(
                    uids=devices,
                    hashcheck='',
                    action=action,
                    del_events=del_events,
                    del_perf=del_perf,
                )
            )
        )

        return response_data

    def _set_components_monitored_by_uids(self, components, monitor=True):
        """
        Sets the monitored state for a list of components.

        Arguments:
            components (list): Uids of the components to update
            monitor (bool): True to monitor, False to stop monitoring

        Returns:
            str: Response message
        """
        response_data = self._router_request(
            self._make_request_data(
                'setComponentsMonitored',
                dict(
                    uids=components,
                    hashcheck='',
                    monitor=monitor,
                )
            )
        )

        return response_data['msg']

    def _lock_components_by_uid(self, components, updates=False, deletion=False,
                                send_event=False):
        """
        Lock devices from changes.

        Arguments:
            components (list): Uids of the components to lock
            updates (bool): Lock components from updates
            deletion (bool): Lock components from deletion
            send_event (bool): Send an event when an action is blocked by locking

        Returns:
            str: Response message
        """
        response_data = self._router_request(
            self._make_request_data(
                'lockComponents',
                dict(
                    uids=components,
                    hashcheck='',
                    updates=updates,
                    deletion=deletion,
                    sendEvent=send_event,
                )
            )
        )

        return response_data['msg']

    def _delete_components_by_uid(self, components):
        """
        Deletes a list of components from a device.

        Arguments:
            components (list): Uids of the components to delete

        Returns:
            str: Response message
        """
        response_data = self._router_request(
            self._make_request_data(
                'deleteComponents',
                dict(
                    uids=components,
                    hashcheck=''
                )
            )
        )

        return response_data['msg']

    def get_tree(self, device_class):
        """
        Get the tree structure of a device class.

        Arguments:
            device_class (str): Device class to use as the top of the tree

        Returns:
            dict:
        """
        tree_data = self._router_request(
            self._make_request_data(
                'getTree',
                dict(id=device_class)
            )
        )

        return self._find_nodes_in_tree(tree_data[0])

    def list_collectors(self):
        """
        Get the list of collectors.

        Returns:
            list:
        """
        return self._router_request(
            self._make_request_data(
                'getCollectors',
                dict(),
            )
        )

    def list_device_classes(self):
        """
        Get the list of all device classes.

        Returns:
            list:
        """
        device_classes_data = self._router_request(
            self._make_request_data(
                'getDeviceClasses',
                dict(),
            )
        )

        dc_list = []
        for dc in device_classes_data['deviceClasses']:
            if len(dc['name']) == 0 or dc['name'] == '/':
                continue
            dc_list.append('Devices{0}'.format(dc['name']))

        return dc_list

    def list_systems(self):
        """
        Get the list of all systems.

        Returns:
            list:
        """
        systems_data = self._router_request(
            self._make_request_data(
                'getSystems',
                dict(),
            )
        )

        systems = []
        for s in systems_data['systems']:
            systems.append(s['name'])

        return systems

    def list_groups(self):
        """
        Get the list of all groups.

        Returns:
            list:
        """
        groups_data = self._router_request(
            self._make_request_data(
                'getGroups',
                dict(),
            )
        )

        groups = []
        for g in groups_data['groups']:
            groups.append(g['name'])

        return groups

    def list_locations(self):
        """
        Get the list of all locations.

        Returns:
            list:
        """
        location_data = self._router_request(
            self._make_request_data(
                'getLocations',
                dict(),
            )
        )

        loci = []
        for l in location_data['locations']:
            loci.append(l['name'])

        return loci

    def get_device_class(self, device_class):
        """
        Get a device class

        Arguments:
            device_class (str): The name of the device class

        Returns:
            ZenossDeviceClass:
        """
        if not device_class.startswith('Devices'):
            if device_class.startswith('/'):
                device_class = 'Devices{0}'.format(device_class)
            else:
                device_class = 'Devices/{0}'.format(device_class)

        dc_data = self._get_info_by_uid(device_class)

        return ZenossDeviceClass(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            dc_data['data']
        )


class ZenossDeviceClass(DeviceRouter):
    """
    Class for Zenoss device class objects
    """

    def __init__(self, url, headers, ssl_verify, device_class_data):
        super(ZenossDeviceClass, self).__init__(url, headers, ssl_verify)

        self.uid = device_class_data['uid'].replace('/zport/dmd/', '', 1)
        self.name = device_class_data['name']
        self.severity = device_class_data['severity']
        self.id = device_class_data['id']
        self.description = device_class_data['description']
        self.connectionInfo = device_class_data['connectionInfo']
        self.events = device_class_data['events']

    def list_devices(self, params=None, keys=None, start=0, limit=50,
                     sort='name', dir='ASC'):
        """
        List the devices contained in a device class. Supports pagination.

        Arguments:
            params (dict): Key/value filters for the search, options are
                name, ipAddress, deviceClass, or productionState
            keys (list): List of keys to return for the devices found
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'name'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'ASC'

        Returns:
            dict:
        """
        if not keys:
            keys = [
                'name',
                'uid',
                'ipAddressString',
                'collector',
                'productionState',
                'priority',
                'location',
                'groups',
                'events'
            ]

        device_data = self._router_request(
            self._make_request_data(
                'getDevices',
                dict(
                    uid=self.uid,
                    params=params,
                    keys=keys,
                    start=start,
                    limit=limit,
                    sort=sort,
                    dir=dir,
                )
            )
        )

        device_list = dict(
            total=device_data['totalCount'],
            hash=device_data['hash'],
            devices=[],
        )
        for device in device_data['devices']:
            devinfo = dict()
            for key in keys:
                if key == "uid":
                    devinfo['uid'] = device['uid'].replace('/zport/dmd/', '', 1)
                elif key == "location":
                    if device['location']:
                        devinfo['location'] = device['location']['name']
                    else:
                        devinfo['location'] = None
                else:
                    devinfo[key] = device[key]
            device_list['devices'].append(devinfo)

        return device_list

    def get_devices(self, params=None, start=0, limit=50, sort='name',
                    dir='ASC'):
        """
        Get the devices contained in a device class. Supports pagination.

        Arguments:
            params (dict): Key/value filters for the search, options are
                name, ipAddress, deviceClass, or productionState
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'name'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'ASC'

        Returns:
            dict(int, str, list(ZenossDevice)): ::

            {
                'total': Total number of devices found
                 'hash': Hashcheck to determine if any devices have changed,
                 'devices': ZenossDevice objects,
            }

        """
        device_data = self._router_request(
            self._make_request_data(
                'getDevices',
                dict(
                    uid=self.uid,
                    params=params,
                    start=start,
                    limit=limit,
                    sort=sort,
                    dir=dir,
                )
            )
        )

        device_list = dict(
            total=device_data['totalCount'],
            hash=device_data['hash'],
            devices=[]
        )
        for device in device_data['devices']:
            device_list['devices'].append(self._get_device_by_uid(
                device['uid'].replace('/zport/dmd/', '', 1)))

        return device_list

    def get_device(self, device_name):
        """
        Get a device from the device class

        Arguments:
            device_name (str): The name of the device to get

        Returns:
            ZenossDevice:
        """
        device_uid = '{0}/devices/{1}'.format(self.uid, device_name)
        return self._get_device_by_uid(device_uid)

    def add_subclass(self, name, description='', connection_info=None):
        """
        Add a new subclass to the device class.

        Arguments:
            name (str): Name of the new subclass
            description (str): Description for the new subclass
            connection_info (list): zProperties that represent the credentials
                for access in the subclass
        """
        return self._add_device_class(name, self.uid, description,
                                      connection_info)

    def add_device(self, device_name, title='', ip_address='', location=None,
                   systems=None, groups=None, model=False,
                   collector='localhost', production_state=500, comments='',
                   priority=3, snmp_community='', snmp_port=161, rack_slot='',
                   hw_manufacturer='', hw_product_name='', os_manufacturer='',
                   os_product_name='', asset_tag='', serial_number='',
                   windows_user='', windows_password='', zcommand_user='',
                   zcommand_password='', configuration_properties=None,
                   custom_properties=None):
        """
        Add a new device to the device class.

        Arguments:
            device_name (str): Name of the new device, will be the device id
            title (str): Optional title for the device, default is to match
                the device_name
            ip_address (str): Ip address for the device, default is to derive
                this from DNS based on device_name
            location (str): Location for the device
            systems (list[(str)]: List of systems for the device
            groups (list[(str)]: List of groups for the device
            model (bool): Set to True to model the device automatically after
                creation
            collector (str): Collector to use for the device
            production_state (int): Numerical production state for the device,
                default is 500 (Pre-Production)
            comments (str): Comments for the device
            priority (int): Numerical priority for the device, default
                is 3 (Normal)
            snmp_community (str): SNMP community string for the device
            snmp_port (int): SNMP port for the device
            rack_slot (str): Rack slot description
            hw_manufacturer (str): Hardware manufacturer name, default is to
                derive by modeling
            hw_product_name (str): Hardware product name, default is to
                derive by modeling
            os_manufacturer (str): Operating system developer, default is to
                derive by modeling
            os_product_name (str): Operating system name, default is to
                derive by modeling
            asset_tag (str): Device's inventory asset tag
            serial_number (str): Device's serial number
            windows_user (str): Username for Windows device monitoring
            windows_password (str): Password for the windows_user
            zcommand_user (str): Username for SSH-based monitoring user
            zcommand_password (str): Password for the zcommand_user
            configuration_properties (dict): Key/value pairs for setting
                Configuration Properties for the device
            custom_properties (dict): Key/value pairs for setting Custom
                Properties for the device

        Returns:
            str: ID of the add device job
        """
        if not configuration_properties:
            configuration_properties = dict()
        if not custom_properties:
            custom_properties = dict()
        if not systems:
            systems = []
        if not groups:
            groups = []

        job_data = self._router_request(
            self._make_request_data(
                'addDevice',
                dict(
                    deviceName=device_name,
                    deviceClass=self.uid.replace('Devices', ''),
                    title=title,
                    snmpCommunity=snmp_community,
                    snmpPort=snmp_port,
                    manageIp=ip_address,
                    locationPath='Locations/{0}'.format(location),
                    systemPaths=systems,
                    groupPaths=groups,
                    model=model,
                    collector=collector,
                    rackSlot=rack_slot,
                    productionState=production_state,
                    comments=comments,
                    hwManufacturer=hw_manufacturer,
                    hwProductName=hw_product_name,
                    osManufacturer=os_manufacturer,
                    osProductName=os_product_name,
                    priority=priority,
                    tag=asset_tag,
                    serialNumber=serial_number,
                    zWinUser=windows_user,
                    zWinPassword=windows_password,
                    zCommandPassword=zcommand_password,
                    zCommandUsername=zcommand_user,
                    zProperties=configuration_properties,
                    cProperties=custom_properties,
                )
            )
        )

        return job_data['new_jobs'][0]['uuid']

    def list_properties(self, params=None, sort=None, sort_dir='ASC'):
        """
        List the configuration properties for the device class

        Arguments:
            params (dict): Search parameters to filter the properties list on.
            sort (str): Sort key for the properties list.
            sort_dir (str): Sort direction, either ASC or DESC

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_properties(self.uid, params=params, sort=sort,
                                  sort_dir=sort_dir)

    def list_local_properties(self):
        """
        List the locally defined configuration properties for the device class

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_local_properties(self.uid)

    def list_custom_properties(self):
        """
        List the custom properties for the device class

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_custom_properties(self.uid)

    def get_properties(self, params=None):
        """
        Get the configuration properties for the device class

        Arguments:
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossProperty objects.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_properties(self.uid, params=params)

    def get_property(self, zproperty):
        """
        Get a configuration property

        Arguments:
            zproperty (str): The id of the property to get

        Returns:
            ZenossProperty:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_property(self.uid, zproperty)

    def get_custom_properties(self, params=None):
        """
        Get the cProperties for the device class

        Arguments:
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossCustomProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossCustomProperty objects.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_custom_properties(self.uid, params=params)

    def get_custom_property(self, cproperty):
        """
        Get a custom property for the device class

        Arguments:
            cproperty (str): ID of the property to get.

        Returns:
            ZenossCustomProperty:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_custom_property(self.uid, cproperty)

    def set_property(self, zproperty, value=None):
        """
        Set the value of a configuration property

        Arguments:
            zproperty (str): The id of the property to set a value for
            value (str): The value to set for the property

        Returns:
            bool:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.set_property_value(self.uid, zproperty, value=value)

    def delete_property(self, zproperty):
        """
        Delete the locally set value of a property for a device class

        Arguments:
            zproperty (str): ID of the property to delete.

        Returns:
            bool:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.delete_property(self.uid, zproperty)


class ZenossDevice(DeviceRouter):
    """
    Class for Zenoss device objects
    """

    def __init__(self, url, headers, ssl_verify, device_data):
        super(ZenossDevice, self).__init__(url, headers, ssl_verify)

        unneeded_props = ['class_label', 'class_plural_label',
                          'class_plural_short_label', 'class_short_label',
                          'deviceClass', 'inspector_type',
                          'ipAddress', 'meta_type', 'priorityLabel',
                          'productionStateLabel', 'pythonClass', 'uuid', ]

        default_props = ['collector', 'comments', 'description', 'device',
                         'deviceConnectionInfo', 'events', 'firstSeen',
                         'groups', 'icon', 'id', 'hwManufacturer', 'hwModel',
                         'lastChanged', 'lastCollected', 'links', 'locking',
                         'memory', 'name', 'osManufacturer', 'osModel',
                         'priority', 'productionState', 'rackSlot',
                         'serialNumber', 'severity', 'snmpAgent',
                         'snmpCommunity', 'snmpContact', 'snmpDescr',
                         'snmpLocation', 'snmpSysName', 'snmpVersion',
                         'sshLink', 'status', 'systems',
                         'tagNumber', 'uptime', ]

        for i in unneeded_props:
            if i in device_data:
                device_data.pop(i)

        uid = device_data.pop('uid')
        self.uid = uid.replace('/zport/dmd/', '', 1)

        location = device_data.pop('location')
        if location:
            self.location = location['name']
        else:
            self.location = None

        self.ip_address = device_data.pop('ipAddressString')

        for prop in default_props:
            setattr(self, prop, device_data.pop(prop))

        self.properties = device_data
        self.parent = self.uid.split('/devices/')[0]

    def list_components(self, meta_type=None, start=0, limit=50, sort='name',
                        dir='ASC', keys=None, name=None):
        """
        Get a list of all the components on a device. Supports pagination.

        Arguments:
            meta_type (str): Meta type of components to list
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'name'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'ASC'
            keys (list): Keys to include in the returned data
            name (str): Regular expression pattern to filter on, requries
                the keys parameter

        Returns:
            dict(int, str, list): ::

            {
                'total': Total number of components found.
                'hash': Hash check to determine if components have changed
                'components': List of components found
            }

        """
        if name and not keys:
            warnings.warn("Filtering by name also requires a list of keys", UserWarning)
            name = None

        components_data = self._router_request(
            self._make_request_data(
                'getComponents',
                dict(
                    uid='/zport/dmd/{0}'.format(self.uid),
                    meta_type=meta_type,
                    start=start,
                    limit=limit,
                    sort=sort,
                    dir=dir,
                    keys=keys,
                    name=name,
                )
            )
        )

        components_list = dict(
            total=components_data['totalCount'],
            hash=components_data['hash'],
            components=[],
        )

        for c in components_data['data']:
            components_list['components'].append(
                c['uid'].split('/{0}/'.format(self.id))[-1])

        return components_list

    def get_components(self, meta_type=None, start=0, limit=50, sort='name',
                       dir='ASC'):
        """
        Get component objects for all components on the device.
        Supports Pagination.

        Arguments:
            meta_type (str): Meta type of components to list
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'name'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'ASC'

        Returns:
            list(ZenossComponent):
        """
        components_list = self.list_components(meta_type=meta_type, start=start,
                                               limit=limit, sort=sort, dir=dir)

        components = []
        for component in components_list['components']:
            c_info = self._get_info_by_uid(
                '{0}/{1}'.format(self.uid, component))
            components.append(
                ZenossComponent(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    c_info['data']
                )
            )

        return components

    def get_component(self, component):
        """
        Get a component object.

        Arguments:
            component (str): Name of the component, e.g. 'hw/cpus/0'

        Returns:
            ZenossComponent:
        """
        c_info = self._get_info_by_uid('{0}/{1}'.format(self.uid, component))

        return ZenossComponent(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            c_info['data']
        )

    def list_user_commands(self):
        """
        Get the list of user commands for a device.

        Returns:
            dict(str, str): ::

            {
                name: Name of the user command
                description: Command description
            }

        """
        uc_data = self._router_request(
            self._make_request_data(
                'getUserCommands',
                dict(uid=self.uid),
            )
        )

        user_commands = []
        for uc in uc_data:
            user_commands.append(
                dict(
                    name=uc['id'],
                    description=uc['description']
                ))

        return user_commands

    def list_local_templates(self):
        """
        Get the list of monitoring templates defined locally on a device.

        Returns:
            list:
        """
        lt_data = self._router_request(
            self._make_request_data(
                'getLocalTemplates',
                dict(
                    uid=self.uid,
                    query='',
                ),
            )
        )

        local_templates = []
        for lt in lt_data['data']:
            local_templates.append(lt['uid'].split('/')[-1])

        return local_templates

    def get_local_templates(self):
        """
        Get ZenossTemplate objects for all locally defined templates.

        Returns:
            list(ZenossTemplate):
        """
        lt_list = self.list_local_templates()

        local_templates = []
        tr = TemplateRouter(self.api_url, self.api_headers, self.ssl_verify)
        for lt in lt_list:
            local_templates.append(
                tr._get_template_by_uid('{0}/{1}'.format(self.uid, lt))
            )

        return local_templates

    def add_local_template(self, template):
        """
        Add a local template to the device.

        Arguments:
            template (str): Name of the new local template
        """
        self._router_request(
            self._make_request_data(
                'addLocalTemplate',
                dict(
                    deviceUid=self.uid,
                    templateId=template,
                )
            )
        )

        return True

    def delete_local_template(self, template):
        """
        Remove a local template from the device.

        Arguments:
            template (str): Name of the template to remove
        """
        self._router_request(
            self._make_request_data(
                'removeLocalTemplate',
                dict(
                    deviceUid=self.uid,
                    templateId=template,
                )
            )
        )

        return True

    def list_active_templates(self):
        """
        Get the list of templates active on a device, both bound and local.

        Returns:
            list(dict(str, str)): ::

            {
                'name': Template name,
                'label': Display label for the template,
            }

        """
        templates_data = self._router_request(
            self._make_request_data(
                'getTemplates',
                dict(id=self.uid),
            )
        )

        templates = []
        for t in templates_data:
            templates.append(dict(
                name=t['uid'].split('/')[-1],
                label=t['text'])
            )

        return templates

    def get_active_templates(self):
        """
        Get ZenossTemplate objects for all active templates on a device.

        Returns:
            list(ZenossTemplate):
        """
        templates_data = self._router_request(
            self._make_request_data(
                'getTemplates',
                dict(id=self.uid),
            )
        )

        templates = []
        tr = TemplateRouter(self.api_url, self.api_headers, self.ssl_verify)
        for t in templates_data:
            tuid = t['uid'].replace('/zport/dmd/', '', 1)
            templates.append(
                tr._get_template_by_uid(tuid)
            )

        return templates

    def list_unbound_templates(self):
        """
        Get the list of available templates that are not bound to the device.

        Returns:
            list(dict(str, str)): ::

            {
                'name': Template name,
                'label': Display label for the template,
            }

        """
        templates_data = self._router_request(
            self._make_request_data(
                'getUnboundTemplates',
                dict(uid=self.uid),
            )
        )

        templates = []
        for t in templates_data['data']:
            templates.append(dict(
                name=t[0],
                label=t[1]
            ))

        return templates

    def get_unbound_templates(self):
        """
        Get ZenossTemplate objects for available templates that are not bound
        to the device.

        Returns:
            list(ZenossTemplate):
        """
        ut_list = self.list_unbound_templates()
        find_path = re.compile('\((\/.*)\)')
        templates = []
        tr = TemplateRouter(self.api_url, self.api_headers, self.ssl_verify)
        for t in ut_list:
            m = re.search(find_path, t['label'])
            if m:
                if m.groups()[0] == "/":
                    path = ''
                else:
                    path = m.groups()[0]
                uid = 'Devices{0}/rrdTemplates/{1}'.format(path, t['name'])
                templates.append(
                    tr._get_template_by_uid(uid)
                )

        return templates

    def list_bound_templates(self):
        """
        Get the list of templates bound to a device, does not include
        local templates.

        Returns:
            list(dict(str, str)): ::

            {
                'name': Template name,
                'label': Display label for the template,
            }

        """
        templates_data = self._router_request(
            self._make_request_data(
                'getBoundTemplates',
                dict(uid=self.uid),
            )
        )

        templates = []
        for t in templates_data['data']:
            templates.append(dict(
                name=t[0],
                label=t[1]
            ))

        return templates

    def get_bound_templates(self):
        """
        Get ZenossTemplate objects templates that are bound to the device.

        Returns:
            list(ZenosTemplate):
        """
        bt_list = self.list_bound_templates()
        find_path = re.compile('\((\/.*)\)')
        templates = []
        tr = TemplateRouter(self.api_url, self.api_headers, self.ssl_verify)
        for t in bt_list:
            m = re.search(find_path, t['label'])
            if m:
                if m.groups()[0] == "/":
                    path = ''
                else:
                    path = m.groups()[0]
                if path.endswith(self.name):
                    path = path.replace(self.name,
                                        'devices/{0}'.format(self.name))
                    uid = 'Devices{0}/{1}'.format(path, t['name'])
                else:
                    uid = 'Devices{0}/rrdTemplates/{1}'.format(path, t['name'])
                templates.append(
                    tr._get_template_by_uid(uid)
                )

        return templates

    def list_overridable_templates(self):
        """
        Get the list of available templates on a device that can be overridden.

        Returns:
            list(dict(str, str)): ::

            {
                'name': Template name,
                'label': Display label for the template,
            }

        """
        template_data = self._router_request(
            self._make_request_data(
                'getOverridableTemplates',
                dict(
                    uid=self.uid,
                    query='',
                ),
            )
        )

        templates = []
        for t in template_data['data']:
            templates.append(dict(
                name=t['uid'].split('/')[-1],
                label=t['label'])
            )

        return templates

    def get_overridable_templates(self):
        """
        Get ZenossTemplate objects for templates that can be overridden.

        Returns:
            list(ZenossTemplate):
        """
        template_data = self._router_request(
            self._make_request_data(
                'getOverridableTemplates',
                dict(
                    uid=self.uid,
                    query='',
                )
            )
        )

        templates = []
        tr = TemplateRouter(self.api_url, self.api_headers, self.ssl_verify)
        for t in template_data['data']:
            templates.append(
                tr._get_template_by_uid(t['uid'].replace('/zport/dmd/', '', 1))
            )

        return templates

    def set_bound_templates(self, templates):
        """
        Set a list of templates as bound to a device.

        Arguments:
            templates (list): List of template names
        """
        self._router_request(
            self._make_request_data(
                'setBoundTemplates',
                dict(
                    uid=self.uid,
                    templateIds=templates,
                )
            )
        )

        return True

    def bind_or_unbind_template(self, path, template):
        """
        Binds a template to the device if it's unbound, or unbinds it if
        it's bound.

        Arguments:
            path (str): Template's path, as given in the display label
            template (str): Name of the template to bind/unbind
        """
        self._router_request(
            self._make_request_data(
                'bindOrUnbindTemplate',
                dict(
                    uid=self.uid,
                    templateUid='Devices{0}/rrdtemplates/{1}'.format(path,
                                                                     template)
                )
            )
        )

        return True

    def reset_bound_templates(self):
        """
        Remove all bound templates from device.
        """
        self._router_request(
            self._make_request_data(
                'resetBoundTemplates',
                dict(uid=self.uid)
            )
        )

        return True

    def move(self, device_class):
        """
        Move the device to a different device class

        Arguments:
            device_class (str): Name of the device class to move the device into

        Returns:
            str: uuid of the Job Manager job for the move
        """
        job_data = self._move_devices_by_uid([self.uid], device_class)
        return job_data[0]['uuid']

    def reidentify(self, new_id):
        """
        Change the device's id in Zenoss. Note that changing the device id will
        cause the loss of all graph data for the device.

        Arguments:
            new_id (str): New ID for the device
        """
        return_data = self._router_request(
            self._make_request_data(
                'renameDevice',
                dict(
                    uid=self.uid,
                    newId=new_id,
                ),
            )
        )

        self.id = new_id
        self.name = new_id
        self.uid = '{0}/devices/{1}'.format(self.parent, self.uid)

        return True

    def lock(self, updates=False, deletion=False, send_event=False):
        """
        Lock the device for changes.

        Arguments:
            updates (bool): Lock for updates
            deletion (bool): Lock for deletion
            send_event (bool): Send an event when an action is blocked by locking

        Returns:
            str: Response message
        """
        return self._lock_devices_by_uid([self.uid], updates=updates,
                                         deletion=deletion,
                                         send_event=send_event)

    def lock_for_updates(self, send_event=False):
        """
        Lock the device for updates.

        Arguments:
            send_event (bool): Send an event when updates are blocked by locking

        Returns:
            str: Response message
        """
        return self.lock(updates=True, send_event=send_event)

    def lock_for_deletion(self, send_event=False):
        """
        Lock the device for updates.

        Arguments:
            send_event (bool): Send an event when deletion is blocked by locking

        Returns:
            str: Response message
        """
        return self.lock(deletion=True, send_event=send_event)

    def reset_ip_address(self, ip_address=''):
        """
        Reset the IP address of the device to ip_address if specified or to the
        result of a DNS lookup if not.

        Arguments:
            ip_address (str): IP address to set device to

        Returns:
            str: Response message
        """
        return self._reset_device_ip_addresses_by_uid([self.uid],
                                                      ip_address=ip_address)

    def set_production_state(self, production_state):
        """
        Set the production state for the device.

        Arguments:
            production_state (int): Numeric value for the desired production
                state.

        Returns:
            str: Response message
        """
        message = self._set_production_state_by_uids([self.uid],
                                                     production_state)
        self.productionState = production_state
        return message

    def set_priority(self, priority):
        """
        Set the priority for the device.

        Arguments:
            priority (int): Numeric value for the desired priority

        Returns:
            str: Reponse message
        """
        message = self._set_priority_by_uids([self.uid], priority)
        self.priority = priority
        return message

    def set_collector(self, collector):
        """
        Set the collector for the device.

        Arguments:
            collector (str): The collector to use for the device

        Returns:
            str: uuid of the Job Manager job for the change
        """
        job_data = self._set_collector_by_uids([self.uid], collector)
        return job_data['uuid']

    def delete(self, action, del_events=False, del_perf=True):
        """
        Remove a device from its organizer, or delete it from Zenoss altogether.

        Arguments:
            action (str): 'remove' to remove the devices from their organizer,
                'delete' to delete them from Zenoss
            del_events (bool): Remove all events for the devices
            del_perf (bool): Remove all perf data for the devices

        Returns:
            bool:
        """
        self._delete_devices_by_uid([self.uid], action, del_events, del_perf)

        return True

    def remodel(self):
        """
        Remodel the device.

        Returns:
            str: uuid of the Job Manager job for the remodel
        """
        response_data = self._router_request(
            self._make_request_data(
                'remodel',
                dict(deviceUid=self.uid)
            )
        )

        return response_data['jobId']

    def list_properties(self, params=None, sort=None, sort_dir='ASC'):
        """
        List the configuration properties for the device

        Arguments:
            params (dict): Search parameters to filter the properties list on.
            sort (str): Sort key for the properties list.
            sort_dir (str): Sort direction, either ASC or DESC

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_properties(self.uid, params=params, sort=sort,
                                  sort_dir=sort_dir)

    def list_local_properties(self):
        """
        List the locally defined configuration properties for the device

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_local_properties(self.uid)

    def list_custom_properties(self):
        """
        List the custom properties for the device

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.list_custom_properties(self.uid)

    def get_properties(self, params=None):
        """
        Get the configuration properties for the device

        Arguments:
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossProperty objects.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_properties(self.uid, params=params)

    def get_property(self, zproperty):
        """
        Get a configuration property

        Arguments:
            zproperty (str): The id of the property to get

        Returns:
            ZenossProperty:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_property(self.uid, zproperty)

    def get_custom_properties(self, params=None):
        """
        Get the cProperties for the device

        Arguments:
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossCustomProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossCustomProperty objects.
            }

        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_custom_properties(self.uid, params=params)

    def get_custom_property(self, cproperty):
        """
        Get a custom property for the device

        Arguments:
            cproperty (str): ID of the property to get.

        Returns:
            ZenossCustomProperty:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.get_custom_property(self.uid, cproperty)

    def set_property(self, zproperty, value=None):
        """
        Set the value of a configuration property

        Arguments:
            zproperty (str): The id of the property to set a value for
            value (str): The value to set for the property

        Returns:
            bool:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.set_property_value(self.uid, zproperty, value=value)

    def delete_property(self, zproperty):
        """
        Delete the locally set value of a property for a device

        Arguments:
            zproperty (str): ID of the property to delete.

        Returns:
            bool:
        """
        pr = PropertiesRouter(self.api_url, self.api_headers, self.ssl_verify)
        return pr.delete_property(self.uid, zproperty)


class ZenossComponent(DeviceRouter):
    """
    Class for Zenoss component objects
    """

    def __init__(self, url, headers, ssl_verify, device_data):
        super(ZenossComponent, self).__init__(url, headers, ssl_verify)

        unneeded_props = ['class_label', 'class_plural_label',
                          'class_plural_short_label', 'class_short_label',
                          'inspector_type', 'uuid']

        for i in unneeded_props:
            if i in device_data:
                device_data.pop(i)

        uid = device_data.pop('uid')
        self.uid = uid.replace('/zport/dmd/', '', 1)
        self.description = device_data.pop('description')
        self.device = device_data.pop('device')
        self.deviceName = device_data.pop('deviceName')
        self.events = device_data.pop('events')
        self.icon = device_data.pop('icon')
        self.id = device_data.pop('id')
        self.locking = device_data.pop('locking')
        self.meta_type = device_data.pop('meta_type')
        self.monitor = device_data.pop('monitor')
        self.monitored = device_data.pop('monitored')
        self.name = device_data.pop('name')
        self.pingStatus = device_data.pop('pingStatus')
        self.severity = device_data.pop('severity')
        self.status = device_data.pop('status')
        self.usesMonitorAttribute = device_data.pop('usesMonitorAttribute')

        self.properties = device_data
        dev_path = self.uid.split('/{0}/'.format(self.device))[0]
        self.parent = '{0}/{1}'.format(dev_path, self.device)

    def set_monitored(self, monitor=True):
        """
        Sets the monitored state for the component.

        Arguments:
            monitor (bool): True to monitor, False to stop monitoring

        Returns:
            str: Response message
        """
        return self._set_components_monitored_by_uids([self.uid], monitor)

    def lock(self, updates=False, deletion=False, send_event=False):
        """
        Lock the component for changes.

        Arguments:
            updates (bool): Lock for updates
            deletion (bool): Lock for deletion
            send_event (bool): Send an event when an action is blocked by locking

        Returns:
            str: Response message
        """
        return self._lock_components_by_uid([self.uid], updates=updates,
                                            deletion=deletion,
                                            send_event=send_event)

    def lock_for_updates(self, send_event=False):
        """
        Lock the component for updates.

        Arguments:
            send_event (bool): Send an event when updates are blocked by locking

        Returns:
            str: Response message
        """
        return self.lock(updates=True, send_event=send_event)

    def lock_for_deletion(self, send_event=False):
        """
        Lock the component for updates.

        Arguments:
            send_event (bool): Send an event when deletion is blocked by locking

        Returns:
            str: Response message
        """
        return self.lock(deletion=True, send_event=send_event)

    def delete(self):
        """
        Delete the component.

        Returns:
            str: Response message
        """
        return self._delete_components_by_uid([self.uid])



# -*- coding: utf-8 -*-

"""
Zenoss jobs_router
"""

from zenossapi.routers import ZenossRouter


class JobsRouter(ZenossRouter):
    """
    Class for interacting with the Zenoss device router
    """

    def __init__(self, url, headers, ssl_verify):
        super(JobsRouter, self).__init__(url, headers, ssl_verify, 'jobs_router', 'JobsRouter')
        self.uuid = None

    def __repr__(self):
        if self.uuid:
            identifier = self.uuid
        else:
            identifier = hex(id(self))

        return '<{0} object at {1}>'.format(
            type(self).__name__, identifier
        )

    def _abort_jobs_by_uuid(self, jobs):
        """
        Aborts the jobs specified in the uuid list.

        Arguments:
            jobs (list): List of job uuids

        Returns:
            bool:
        """
        jobs_abort_response = self._router_request(
            self._make_request_data(
                'abort',
                dict(jobids=jobs),
            )
        )

        return True

    def _delete_jobs_by_uuid(self, jobs):
        """
        Deletes the jobs specified in the uuid list.

        Arguments:
            jobs (list): List of jobs uuids

        Returns:
            list: List of uuids deleted
        """
        deletes_response = self._router_request(
            self._make_request_data(
                'deleteJobs',
                dict(
                    jobids=jobs
                )
            )
        )

        if deletes_response is None:
            return []
        else:
            return deletes_response['deletedJobs']

    def list_jobs(self, start=0, limit=50, sort='scheduled', dir='DESC'):
        """
        List all Job Manager jobs, supports pagination.

        Arguments:
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'scheduled'. Other
                sort keys are 'started, 'finished', 'status', 'type' and 'user'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'DESC'

        Returns:
            dict(int, dict(str, int, int, int, str, str, str, str, str)): ::

            {
                'total': (int) Total number of jobs,
                'jobs': {
                    'description': (str) Job description,
                    'finished': (int) Time the job finished in timestamp format,
                    'scheduled': (int) Time the job was scheduled in timestamp format,
                    'started': (int) Time the job started in timestamp format,
                    'status': (str) Status of the job,
                    'type': (str) Job type,
                    'uid': (str) JobManager UID - /zport/dmd/JobManager,
                    'user': (str) User who scheduled the job,
                    'uuid': (str) UUID of the job,
                }
            }

        """
        jobs_data = self._router_request(
            self._make_request_data(
                'getJobs',
                dict(
                    start=start,
                    limit=limit,
                    sort=sort,
                    dir=dir,
                    page=0,
                )
            )
        )

        return dict(
            total=jobs_data['totalCount'],
            jobs=jobs_data['jobs'],
        )

    def get_jobs(self, start=0, limit=50, sort='scheduled', dir='ASC'):
        """
        Get ZenossJob objects for Job Manager jobs. Supports pagination.

        Arguments:
            start (int): Offset to start device list from, default 0
            limit (int): The number of results to return, default 50
            sort (str): Sort key for the list, default is 'scheduled'. Other
                sort keys are 'started, 'finished', 'status', 'type' and 'user'
            dir (str): Sort order, either 'ASC' or 'DESC', default is 'ASC'

        Returns:
            list(ZenossJob):
        """
        jobs_data = self.list_jobs(start=start, limit=limit, sort=sort, dir=dir)

        jobs = []
        for job in jobs_data['jobs']:
            jobs.append(self.get_job(job['uuid']))

        return jobs

    def get_job(self, job):
        """
        Get a ZenossJob object by the job's uuid

        Arguments:
            job (str): uuid of the job

        Returns:
            ZenossJob:
        """
        job_data = self._router_request(
            self._make_request_data(
                'getInfo',
                dict(jobid=job)
            )
        )

        return ZenossJob(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            job_data['data']
        )


class ZenossJob(JobsRouter):
    """
    Class for Zenoss job objects
    """

    def __init__(self, url, headers, ssl_verify, job_data):
        super(ZenossJob, self).__init__(url, headers, ssl_verify)

        self.description = job_data['description']
        self.duration = job_data['duration']
        self.errors = job_data['errors']
        self.finished = job_data['finished']
        self.id = job_data['id']
        self.logfile = job_data['logfile']
        self.meta_type = job_data['meta_type']
        self.name = job_data['name']
        self.scheduled = job_data['scheduled']
        self.started = job_data['started']
        self.status = job_data['status']
        self.type = job_data['type']
        self.uid = job_data['uid']
        self.user = job_data['user']
        self.uuid = job_data['uuid']

    def abort(self):
        """
        Abort the job.

        Returns:
            bool:
        """
        return self._abort_jobs_by_uuid([self.uuid])

    def delete(self):
        """
        Delete the job.

        Returns:
            list: Job ID
        """
        return self._delete_jobs_by_uuid([self.uuid])

    def get_log(self):
        """
        Get the log for the job.

        Returns:
            dict(str, bool, list): ::

            {
                'logfile': Filesystem path of the log file,
                'maxLimit': True or False,
                'content': Log file lines
            }

        """
        return self._router_request(
            self._make_request_data(
                'detail',
                dict(
                    jobid=self.uuid,
                )
            )
        )




# -*- coding: utf-8 -*-

"""
Zenoss evconsole_router
"""

from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter


class EventsRouter(ZenossRouter):
    """
    Class for interacting with Zenoss events.
    """

    def __init__(self, url, headers, ssl_verify):
        super(EventsRouter, self).__init__(url, headers, ssl_verify,
                                           'evconsole_router', 'EventsRouter')

        self.evid = None
        self.event_states_map = dict(
            New=0,
            Acknowledged=1,
            Suppressed=2,
            Closed=3,
            Cleared=4,
            Aged=6,
        )
        self.event_severity_map = dict(
            Critical=5,
            Error=4,
            Warning=3,
            Info=2,
            Debug=1,
            Clear=0,
        )

    def __repr__(self):
        if self.evid:
            identifier = "id {0}".format(self.evid)
        else:
            identifier = "at {0}".format(hex(id(self)))

        return '<{0} object {1}>'.format(
            type(self).__name__, identifier
        )

    def _query_events(self, limit=10, start=0, sort='lastTime', sort_dir='DESC',
                      params=None, exclude=None, keys=None,
                      context=None, archive=False):
        """
        Get a list of events based on query parameters.

        Some search parameter examples:

        Find New events:
            params=dict(eventState=[0])
        Find open events (in New or Acknowledged state):
            params=dict(eventState=[0,1])
        Find open events for Production servers:
            params=dict(eventState=[0,1], prodState=[1000])

        Other handy search keys: agent, count, DevicePriority, firstTime,
        lastTime, monitor, message, severity, summary

        Arguments:
            limit (int): Maximum number of events to get
            start (int): Minimum index of events to get
            sort (str): Sort key for events list
            sort_dir (str): Sort direction, ASC or DESC
            params (dict): Key/value pairs for filtering the events to include
            exclude (dict): Key/value pairs to filter out of the found events
            keys (list): List of keys to include in the return event data
            context (str): Device or device class uid as context for the query
            archive (bool): Search the events archive instead of active events

        Returns:
            dict(int, float, list(dict)): ::

            {
                'total': Total count of events matched by the query,
                'ts': Timestamp for the search,
                'events': List of events found,
            }

        """
        events_data = self._router_request(
            self._make_request_data(
                'query',
                dict(
                    limit=limit,
                    start=start,
                    sort=sort,
                    dir=sort_dir,
                    params=params,
                    exclusion_filter=exclude,
                    keys=keys,
                    uid=context,
                    archive=archive,
                )
            )
        )

        return dict(
            ts=events_data['asof'],
            total=events_data['totalCount'],
            events=events_data['events'],
        )

    def _get_details_by_evid(self, evid):
        """
        Get the details of an event by the event ID.

        Arguments:
            evid (str): The event ID

        Returns:
            dict: Event details
        """
        event_data = self._router_request(
            self._make_request_data(
                'detail',
                dict(
                    evid=evid,
                )
            )
        )

        return event_data['event'][0]

    def _event_actions(self, action, evids=None, exclude_ids=None, params=None,
                      context=None, last_change=None, limit=None, timeout=60):
        """
        Close events by either passing a list of event ids or by a query.

        Arguments:
            action (str): The action to take on the event - close,
                ack, or reopen
            evids (list): List of event IDs to close
            exclude_ids (list): List of event IDs to exclude from the close
            params (dict): Key/value pairs for filtering the events to close
            context (str): Device or device class uid as context for the query
            last_change (float): Timestamp - only close events if there has
                been no change since this time
            limit (int): Maximum number of events to return in the query
            timeout (int): Number of seconds before the query times out
        """
        if action not in ['close', 'acknowledge', 'reopen']:
            raise ZenossAPIClientError("Unknown event action: {0}".format(action))

        self._router_request(
            self._make_request_data(
                action,
                dict(
                    evids=evids,
                    excludeIds=exclude_ids,
                    params=params,
                    uid=context,
                    asof=last_change,
                    limit=limit,
                    timeout=timeout,
                )
            )
        )

        return True

    def get_config(self):
        """
        Get the event handling configuration.

        Returns:
            list(dict):
        """
        config_data = self._router_request(
            self._make_request_data(
                'getConfig',
                dict(),
            )
        )

        return config_data['data']

    def update_config(self, config_values):
        """
        Update the Zenoss event handling configuration.

        Arguments:
            config_values (dict): Key/value pairs of the config
                values to change.
        """
        self._router_request(
            self._make_request_data(
                'setConfigValues',
                dict(values=config_values),
            )
        )

        return True

    def get_event_by_evid(self, evid):
        """
        Get an event by its event id

        Arguments:
            evid (str): The event id

        Returns:
            ZenossEvent:
        """
        event_data = self._get_details_by_evid(evid)

        return ZenossEvent(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            event_data,
        )

    def list_open_events(self, limit=10, start=0, sort='lastTime', sort_dir='DESC'):
        """
        Get a list of all open events (new or acknowledged state)

        Arguments:
            limit (int): Maximum number of events to return
            start (int): Minimum index of events to get
            sort (str): Sort key for events list
            sort_dir (str): Sort direction, ASC or DESC

        Returns:
            dict:
        """
        return self._query_events(
            limit=limit,
            start=start,
            sort=sort,
            sort_dir=sort_dir,
            params=dict(eventState=[0, 1], severity=[3, 4, 5])
        )

    def get_open_events(self, limit=10, start=0, sort='lastTime', sort_dir='DESC'):
        """
        Get all open events (new or acknowledged state)

        Arguments:
            limit (int): Maximum number of events to return
            start (int): Minimum index of events to get
            sort (str): Sort key for events list
            sort_dir (str): Sort direction, ASC or DESC

        Returns:
            list(ZenossEvent):
        """
        events_data = self._query_events(
            limit=limit,
            start=start,
            sort=sort,
            sort_dir=sort_dir,
            params=dict(eventState=[0, 1], severity=[3, 4, 5]),
            keys=['evid']
        )

        events = []
        for e in events_data['events']:
            evdetails = self._get_details_by_evid(e['evid'])
            events.append(
                ZenossEvent(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    evdetails,
                )
            )

        return events

    def list_open_production_events(self, limit=10, start=0, sort='lastTime',
                                    sort_dir='DESC'):
        """
        Get a list of all open events (new or acknowledged state) for devices
        with a production state of Production

        Arguments:
            limit (int): Maximum number of events to return
            start (int): Minimum index of events to get
            sort (str): Sort key for events list
            sort_dir (str): Sort direction, ASC or DESC

        Returns:
            dict:
        """
        return self._query_events(
            limit=limit,
            start=start,
            sort=sort,
            sort_dir=sort_dir,
            params=dict(eventState=[0, 1], prodState=[1000], severity=[3, 4, 5])
        )

    def get_open_production_events(self, limit=10, start=0, sort='lastTime',
                                   sort_dir='DESC'):
        """
        Get all open events (new or acknowledged state) for devices with a
        production state of Production

        Arguments:
            limit (int): Maximum number of events to return
            start (int): Minimum index of events to get
            sort (str): Sort key for events list
            sort_dir (str): Sort direction, ASC or DESC

        Returns:
            list(ZenossEvent):
        """
        events_data = self._query_events(
            limit=limit,
            start=start,
            sort=sort,
            sort_dir=sort_dir,
            params=dict(eventState=[0, 1], prodState=[1000], severity=[3, 4, 5]),
            keys=['evid']
        )

        events = []
        for e in events_data['events']:
            evdetails = self._get_details_by_evid(e['evid'])
            events.append(
                ZenossEvent(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    evdetails,
                )
            )

        return events

    def add_event(self, summary, device, severity, component=None,
                  event_class_key=None, event_class='/Status'):
        """
        Create a new Zenoss event.

        Arguments:
            summary (str): Summary for the new event
            device (str): Device ID for the new event
            component (str): Component UID for the new event
            severity (str): Severity to assign the new event, must be one of
                Critical, Error, Warning, Info, Debug, or Clear
            event_class_key (str): The Event Class Key to assign to the event
            event_class (str): Event Class for the event

        Returns:
            ZenossEvent:
        """
        if severity in self.event_severity_map:
            severity_num = self.event_severity_map[severity]
        else:
            severity_num = severity
        self._router_request(
            self._make_request_data(
                'add_event',
                dict(
                    summary=summary,
                    device=device,
                    component=component,
                    severity=severity,
                    evclasskey=event_class_key,
                    evclass=event_class,
                )
            )
        )

        filter = dict(
            summary=summary,
            device=[device],
            component=[component] if component else None,
            severity=severity_num,
            eventClassKey=event_class_key,
        )

        event_data = self._query_events(
            limit=1,
            params=filter,
            keys=['evid']
        )

        return self.get_event_by_evid(event_data['events'][0]['evid'])

    def clear_heartbeats(self):
        """
        Clear all heartbeat events

        Returns:
            bool: True on success
        """
        self._router_request(
            self._make_request_data(
                'clear_heartbeats',
                dict(),
            )
        )

        return True

    def clear_heartbeat(self, collector, daemon):
        """
        Clear a heartbeat event for a specific daemon.

        Arguments:
            collector (str): Collector the daemon is running in,
                e.g. slvcollector
            daemon (str): Monitoring daemon to clear the heartbeat event
                for, e.g. zencommand
        """
        self._router_request(
            self._make_request_data(
                'clear_heartbeat',
                dict(
                    monitor=collector,
                    daemon=daemon,
                )
            )
        )

        return True


class ZenossEvent(EventsRouter):
    """
    Class for Zenoss event objects
    """

    def __init__(self, url, headers, ssl_verify, event_data):
        super(ZenossEvent, self).__init__(url, headers, ssl_verify)

        self.evid = event_data['evid']
        self.device_class = event_data['DeviceClass']
        self.device_groups = event_data['DeviceGroups']
        self.device_priority = event_data['DevicePriority']
        self.location = event_data['Location']
        self.systems = event_data['Systems']
        self.agent = event_data['agent']
        self.clearid = event_data['clearid']
        self.component = event_data['component']
        self.component_title = event_data['component_title']
        self.component_url = event_data['component_url']
        self.component_uuid = event_data['component_uuid']
        self.count = event_data['count']
        self.dedupid = event_data['dedupid']
        self.details = event_data['details']
        self.device = event_data['device']
        self.device_title = event_data['device_title']
        self.device_url = event_data['device_url']
        self.device_uuid = event_data['device_uuid']
        self.event_class = event_data['eventClass']
        self.event_class_key = event_data['eventClassKey']
        self.event_class_mapping = event_data['eventClassMapping']
        self.event_class_mapping_url = event_data['eventClassMapping_url']
        self.event_class_url = event_data['eventClass_url']
        self.event_group = event_data['eventGroup']
        self.event_key = event_data['eventKey']
        self.event_state = event_data['eventState']
        self.facility = event_data['facility']
        self.first_time = event_data['firstTime']
        self.id = event_data['id']
        self.ip_address = event_data['ipAddress']
        self.last_time = event_data['lastTime']
        self.log = event_data['log']
        self.message = event_data['message']
        self.collector = event_data['monitor']
        self.ntevid = event_data['ntevid']
        self.ownerid = event_data['ownerid']
        self.priority = event_data['priority']
        self.production_state = event_data['prodState']
        self.severity = event_data['severity']
        self.state_change = event_data['stateChange']
        self.summary = event_data['summary']

    def update_log(self, message):
        """
        Add an entry to the event's log

        Arguments:
            message (str): Log entry to add
        """
        self._router_request(
            self._make_request_data(
                'write_log',
                dict(
                    evid=self.evid,
                    message=message,
                )
            )
        )

        event_detail = self._get_details_by_evid(self.evid)
        self.log = event_detail['log']

        return True

    def close(self):
        """
        Close the event.
        """
        return self._event_actions('close', [self.evid])

    def ack(self):
        """
        Acknowledge the event.
        """
        return self._event_actions('acknowledge', [self.evid])

    def reopen(self):
        """
        Reopen (unacknowledge or unclose) the event.
        """
        return self._event_actions('reopen', [self.evid])



# -*- coding: utf-8 -*-

"""
Zenoss properties_router
"""

from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter


class PropertiesRouter(ZenossRouter):
    """
    Class for interacting with Zenoss properties.
    """

    def __init__(self, url, headers, ssl_verify):
        super(PropertiesRouter, self).__init__(url, headers, ssl_verify,
                                               'properties_router',
                                               'PropertiesRouter')

        self.id = None

    def __repr__(self):
        if self.id:
            identifier = "id {0}".format(self.id)
        else:
            identifier = "at {0}".format(hex(id(self)))

        return '<{0} object {1}>'.format(
            type(self).__name__, identifier
        )

    def _refresh(self, property_data):
        """
        Refresh the attributes of the object, e.g. after deleting a local copy.

        Arguments:
            property_data (dict): New attribute values
        """
        skip = ['api_action', 'api_endpoint', 'api_headers', 'api_url', 'ssl_verify']

        for pd in property_data:
            if pd in skip:
                continue
            setattr(self, pd, property_data[pd])

    def list_properties(self, uid, params=None, sort=None, sort_dir='ASC'):
        """
        Get a list of ZenProperties for the uid context.

        Arguments:
            uid (str): UID of the object to list properties for.
            params (dict): Search parameters to filter the properties list on.
            sort (str): Sort key for the properties list.
            sort_dir (str): Sort direction, either ASC or DESC

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        properties_data = self._router_request(
            self._make_request_data(
                'getZenProperties',
                dict(
                    uid=uid,
                    params=params,
                    sort=sort,
                    dir=sort_dir,
                )
            )
        )

        return dict(
            total=properties_data['totalCount'],
            properties=properties_data['data'],
        )

    def get_properties(self, uid, params=None):
        """
        Get ZenossProperties objects for the properties of a uid context.

        Arguments:
            uid (str): UID of the object to get properties for.
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossProperty objects.
            }

        """
        properties_data = self.list_properties(uid, params=params)

        props = dict(
            total=properties_data['total'],
            properties=[]
        )
        for prop in properties_data['properties']:
            props['properties'].append(
                ZenossProperty(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    prop,
                )
            )

        return props

    def list_local_properties(self, uid):
        """
        Get a list of properties set locally to the specified UID.

        Arguments:
            uid (str): UID to get local properties for.

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }
        """
        return self.list_properties(uid, params=dict(islocal='1'))

    def get_local_properties(self, uid):
        """
        Get ZenossProperty objects for the local properties of a specified uid.

        Arguments:
            uid (str): UID to get local properties for.

        Returns:
            dict(int, list(ZenossProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossProperty objects.
            }

        """
        properties_data = self.list_local_properties(uid)

        props = dict(
            total=properties_data['total'],
            properties=[]
        )
        for prop in properties_data['properties']:
            props['properties'].append(
                ZenossProperty(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    prop,
                )
            )

        return props

    def get_property(self, uid, zproperty):
        """
        Get a single ZenossProperty

        Arguments:
            uid (str): UID to get the property of.
            zproperty (str): ID of the property to get.

        Returns:
            ZenossProperty:
        """
        prop_data = self._router_request(
            self._make_request_data(
                'getZenProperties',
                dict(
                    uid=uid,
                    params=dict(
                        id=zproperty,
                    )
                )
            )
        )

        return ZenossProperty(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            prop_data['data'][0]
        )

    def list_custom_properties(self, uid, params=None, sort=None, sort_dir='ASC'):
        """
        Get a list of cProperties for the uid context.

        Arguments:
            uid (str): UID of the object to list properties for.
            params (dict): Search parameters to filter the properties list on.
            sort (str): Sort key for the properties list.
            sort_dir (str): Sort direction, either ASC or DESC

        Returns:
            dict(int, list(dict)): ::

            {
                'total': Total count of properties returned.
                'properties': List of properties found.
            }

        """
        properties_data = self._router_request(
            self._make_request_data(
                'getCustomProperties',
                dict(
                    uid=uid,
                    params=params,
                    sort=sort,
                    dir=sort_dir,
                )
            )
        )

        return dict(
            total=properties_data['totalCount'],
            properties=properties_data['data'],
        )

    def get_custom_properties(self, uid, params=None):
        """
        Get ZenossCustomProperties objects for the cProperties of a uid context.

        Arguments:
            uid (str): UID of the object to get properties for.
            params (dict): Search parameters for filter the properties on.

        Returns:
            dict(int, list(ZenossCustomProperty)): ::

            {
                'total': Total count of properties returned.
                'properties': List of ZenossCustomProperty objects.
            }

        """
        properties_data = self.list_custom_properties(uid, params=params)

        props = dict(
            total=properties_data['total'],
            properties=[]
        )
        for prop in properties_data['properties']:
            props['properties'].append(
                ZenossCustomProperty(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    prop,
                )
            )

        return props

    def get_custom_property(self, uid, cproperty):
        """
        Get a single ZenossCustomProperty

        Arguments:
            uid (str): UID to get the property of.
            cproperty (str): ID of the property to get.

        Returns:
            ZenossCustomProperty:
        """
        prop_data = self._router_request(
            self._make_request_data(
                'getCustomProperties',
                dict(
                    uid=uid,
                    params=dict(
                        id=cproperty,
                    )
                )
            )
        )

        return ZenossCustomProperty(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            prop_data['data'][0]
        )

    def set_property_value(self, uid, zproperty, value=None):
        """
        Sets (or updates) the local value of a property

        Arguments:
            value: The new value for the property, type varies by property.

        Returns:
            bool:
        """
        property_data = self._router_request(
            self._make_request_data(
                'setZenProperty',
                dict(
                    uid=uid,
                    zProperty=zproperty,
                    value=value,
                )
            )
        )

        return property_data['data']

    def delete_property(self, uid, zproperty):
        """
        Delete a ZenProperty.

        Arguments:
            uid (str): UID to delete the property from
            zproperty (str): ID of the property to delete.

        Returns:
            bool:
        """
        self._router_request(
            self._make_request_data(
                'deleteZenProperty',
                dict(
                    uid=uid,
                    zProperty=zproperty
                )
            )
        )

        return True


class ZenossProperty(PropertiesRouter):
    """
    Class for ZenProperties
    """

    def __init__(self, url, headers, ssl_verify, property_data):
        super(ZenossProperty, self).__init__(url, headers, ssl_verify)

        self.id = property_data['id']
        self.category = property_data['category']
        self.description = property_data['description']
        self.islocal = property_data['islocal']
        self.label = property_data['label']
        self.value = property_data['value']
        self.valueAsString = property_data['valueAsString']
        self.path = 'Devices{0}'.format(property_data['path'])
        self.type = property_data['type']
        self.options = []

    def set_value(self, path=None, value=None):
        """
        Sets (or updates) the local value of a property

        Arguments:
            path (str): UID of the node to set the property for.
            value (str): The new value for the property, type varies by property.

        Returns:
            bool:
        """
        if path:
            self.path = path

        property_data = self.set_property_value(self.path, self.id, value=value)

        self.value = property_data['value']
        self.valueAsString = property_data['valueAsString']
        self.islocal = '1'

        return True

    def delete(self):
        """
        Delete the local instance of a property.

        Returns:
            bool:
        """
        if self.islocal:
            self.delete_property(self.path, self.id)
            defprop = self.list_properties(self.path, params=dict(id=self.id))
            self._refresh(defprop['properties'][0])
            return True

        return False


class ZenossCustomProperty(PropertiesRouter):
    """
    Class for Zenoss CustomProperties
    """

    def __init__(self, url, headers, ssl_verify, property_data):
        super(ZenossCustomProperty, self).__init__(url, headers, ssl_verify)

        self.id = property_data['id']
        self.islocal = property_data['islocal']
        self.label = property_data['label']
        self.value = property_data['value']
        self.valueAsString = property_data['valueAsString']
        self.path = 'Devices{0}'.format(property_data['path'])
        self.type = property_data['type']
        self.options = property_data['options']

    def set_value(self, path=None, value=None):
        """
        Sets (or updates) the local value of a custom property

        Arguments:
            path (str): UID of the node to set the property for.
            value (str): The new value for the property, type varies by property.

        Returns:
            bool:
        """
        if path:
            self.path = path

        property_data = self.set_property_value(self.path, self.id, value=value)

        self.value = property_data['value']
        self.valueAsString = property_data['valueAsString']
        self.islocal = '1'

        return True

    def delete(self):
        """
        Delete the local instance of a property.

        Returns:
            bool:
        """
        if self.islocal:
            self.delete_property(self.path, self.id)
            defprop = self.list_custom_properties(self.path, params=dict(id=self.id))
            self._refresh(defprop['properties'][0])
            return True

        return False



# -*- coding: utf-8 -*-

import requests
from zenossapi.apiclient import ZenossAPIClientError, ZenossAPIClientAuthenticationError


class ZenossRouter(object):
    """
    Base class for Zenoss router classes
    """
    def __init__(self, url, headers, ssl_verify, endpoint, action):
        self.api_url = url
        self.api_headers = headers
        self.ssl_verify = ssl_verify
        self.api_endpoint = endpoint
        self.api_action = action

    def _check_uid(self, uid):
        if not uid.startswith('Devices'):
            if uid.startswith('/'):
                uid = 'Devices{0}'.format(uid)
            else:
                uid = 'Devices/{0}'.format(uid)

        return uid

    def _make_request_data(self, method, data):
        return dict(
            action=self.api_action,
            method=method,
            data=[data],
            tid=1,
        )

    def _router_request(self, data):
        # Disable warnings from urllib3 if ssl_verify is False, otherwise
        # every request will print an InsecureRequestWarning
        if not self.ssl_verify:
            requests.urllib3.disable_warnings()

        response = requests.post(
            '{0}/{1}'.format(self.api_url, self.api_endpoint),
            headers=self.api_headers,
            json=data,
            verify=self.ssl_verify,
        )

        if response.ok:
            if response.url.find('login_form') > -1:
                raise ZenossAPIClientAuthenticationError('API Login Failed')
            response_json = response.json()
            if 'result' in response_json:
                if response_json['result']:
                    if 'success' in response_json['result']:
                        if not response_json['result']['success']:
                            raise ZenossAPIClientError('Request failed: {}'.format(response_json['result']['msg']))
            else:
                raise ZenossAPIClientError('Request failed, no response data returned!')

            return response_json['result']

        else:
            raise ZenossAPIClientError('Request failed: {0} {1}'.format(response.status_code, response.reason))



# -*- coding: utf-8 -*-

"""
Zenoss template_router
"""

from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter


class TemplateRouter(ZenossRouter):
    """
    Class for interacting with the Zenoss template router
    """

    def __init__(self, url, headers, ssl_verify):
        super(TemplateRouter, self).__init__(url, headers, ssl_verify, 'template_router', 'TemplateRouter')
        self.uid = None
        self.properties = None

    def __repr__(self):
        if self.uid:
            identifier = self.uid
        else:
            identifier = hex(id(self))

        return '<{0} object at {1}>'.format(
            type(self).__name__, identifier
        )

    def _get_properties(self, zobject):
        """
        Gets the properties of an object.

        Arguments:
            zobject (str): The uid of the Zenoss object (device, component,
                           etc.) to get the properties of

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'getInfo',
                dict(uid=zobject)
            )
        )

    def _get_template_by_uid(self, template_uid):
        """
        Gets a template by its full UID

        Arguments:
            template_uid (str): UID of the template

        Returns:
            ZenossTemplate:
        """
        template_data = self._router_request(
            self._make_request_data(
                'getInfo',
                dict(uid=template_uid)
            )
        )
        return ZenossTemplate(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            template_data['data']
        )

    def _get_data_source_by_uid(self, datasource_uid):
        """
        Get a data source by its full UID.

        Arguments:
            datasource_uid (str): UID of the data source to get

        Returns:
            ZenossDataSource:
        """
        data_source_data = self._router_request(
            self._make_request_data(
                'getDataSourceDetails',
                dict(uid=datasource_uid),
            )
        )
        return ZenossDataSource(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            data_source_data['record']
        )

    def _get_data_point_by_uid(self, datapoint_uid):
        """
        Get a data point by its full UID.

        Arguments:
            datapoint_uid (str): UID of the data point to get details for

        Returns:
            ZenossDataPoint:
        """
        dp_data = self._router_request(
            self._make_request_data(
                'getDataPointDetails',
                dict(uid=datapoint_uid),
            )
        )
        return ZenossDataPoint(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            dp_data['record']
        )

    def _get_threshold_by_uid(self, threshold_uid):
        """
        Gets a threshold by its full UID

        Arguments:
            threshold_uid (str): UID of the template

        Returns:
            ZenossThreshold:
        """
        threshold_data = self._router_request(
            self._make_request_data(
                'getThresholdDetails',
                dict(uid=threshold_uid)
            )
        )

        return ZenossThreshold(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            threshold_data['record']
        )

    def _get_graph_by_uid(self, graph_uid):
        """
        Get a graph by its full UID.

        Arguments:
            graph_uid (str): UID of the graph to get the definition of

        Returns:
            ZenossGraph:
        """
        graph_data = self._router_request(
            self._make_request_data(
                'getGraphDefinition',
                dict(uid=graph_uid),
            )
        )

        return ZenossGraph(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            graph_data['data']
        )

    def _find_templates_in_tree(self, templates_data):
        """
        Works through the dict structure returned by the Zenoss API for
        template queries and returns the defined templates from it.

        Arguments:
            templates_data (dict): Templates data returned by the API

        Returns:
            list:
        """
        templates = []
        for node in templates_data['children']:
            if node['leaf']:
                if node['text'].find("Locally Defined") > -1:
                    templates.append((node['uid'].replace('/zport/dmd/', '', 1), node['qtip']))
            else:
                templates.extend(self._find_templates_in_tree(node))

        return templates

    def _set_properties(self, properties):
        """
        Sets arbitrary properties of any object.

        Arguments:
            properties (dict): Properties and values to set

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'setInfo',
                properties
            )
        )

    def set_properties(self, properties):
        """
        Sets properties of an object.

        Arguments:
            properties (dict): Properties and values to set
        """
        if not isinstance(properties, dict):
            raise ZenossAPIClientError('Type error: Properties to set for {} must be a dict'.format(type(self).__name__))

        if not self.uid:
            return

        data = properties
        data['uid'] = self.uid
        properties_result = self._set_properties(data)
        for prop in properties:
            if getattr(self, prop, False):
                setattr(self, prop, properties[prop])
            elif getattr(self, 'properties', False) and prop in self.properties:
                self.properties[prop] = properties[prop]

        return properties_result

    def get_all_templates(self):
        """
        Returns all defined templates.

        Returns:
            list(ZenossTemplate):
        """
        templates_data = self._router_request(
            self.get_device_class_templates(
                device_class='Devices'
            )
        )

        templates = []
        found_templates = self._find_templates_in_tree(templates_data[0])
        for t in found_templates:
            templates.append(
                self._get_template_by_uid(t[0])
            )
        return templates

    def list_all_templates(self):
        """
        Returns all defined templates as a list of tuples containing the
        template UID and description.

        Returns:
            list(ZenossTemplate):
        """
        templates_data = self._router_request(
            self.get_device_class_templates(
                device_class='Devices'
            )
        )

        templates = []
        found_templates = self._find_templates_in_tree(templates_data[0])
        for t in found_templates:
            templates.append(t)
        return templates

    def get_device_class_templates(self, device_class):
        """
        Gets the defined templates for a device class

        Arguments:
            device_class (str): Device class to get templates for

        Returns:
            list(ZenossTemplate):
        """
        templates_data = self._router_request(
            self._make_request_data(
                'getDeviceClassTemplates',
                dict(id=device_class),
            )
        )

        templates = []
        found_templates = self._find_templates_in_tree(templates_data[0])
        for t in found_templates:
            templates.append(
                self._get_template_by_uid(t[0])
            )
        return templates

    def list_device_class_templates(self, device_class):
        """
        Returns the defined templates for a device class as a list of
        tuples containing the template UID and description.

        Arguments:
            device_class (str): Device class to list templates for

        Returns:
            list(str):
        """
        if not device_class.startswith('Devices'):
            if device_class.startswith('/'):
                device_class = 'Devices{0}'.format(device_class)
            else:
                device_class = 'Devices/{0}'.format(device_class)
        templates_data = self._router_request(
            self._make_request_data(
                'getDeviceClassTemplates',
                dict(id=device_class),
            )
        )

        templates = []
        found_templates = self._find_templates_in_tree(templates_data[0])
        for t in found_templates:
            templates.append(t)
        return templates

    def get_object_templates(self, zenoss_object):
        """
        Gets the templates bound to a specific object
        (monitored resource or component)

        Arguments:
            zenoss_object (str): The uid of the object, e.g.
                        Devices/Server/Zuora/Aspose/devices/10.aspose.prod.slv.zuora

        Returns:
            list(ZenossTemplate):
        """
        if not zenoss_object.startswith('Devices'):
            if zenoss_object.startswith('/'):
                zenoss_object = 'Devices{0}'.format(zenoss_object)
            else:
                zenoss_object = 'Devices/{0}'.format(zenoss_object)

        templates_data = self._router_request(
            self._make_request_data(
                'getObjTemplates',
                dict(uid=zenoss_object),
            )
        )

        templates = []
        found_templates = templates_data['data']
        for t in found_templates:
            templates.append(
                self._get_template_by_uid(t['uid'].replace('/zport/dmd/', '', 1))
            )
        return templates

    def get_template(self, device_class, template):
        """
        Get a Zenoss template

        Arguments:
            device_class (str): Name of the device class where the template is defined
            template (str): Name of the template to get

        Returns:
            ZenossTemplate:
        """
        # Check to see if this is a local template instead of a device class
        # template
        if "devices" in device_class:
            template_uid = '{0}/{1}'.format(device_class, template)
        else:
            template_uid = '{0}/rrdTemplates/{1}'.format(device_class, template)

        return self._get_template_by_uid(template_uid)

    def add_template(self, target, name):
        """
        Adds a template to a device class.

        Arguments:
            target (str): The uid of the target device class
            name (str): Unique name of the template to add

        Returns:
            ZenossTemplate:
        """
        if not target.startswith('Devices'):
            if target.startswith('/'):
                target = 'Devices{0}'.format(target)
            else:
                target = 'Devices/{0}'.format(target)

        if not target.endswith('rrdTemplates'):
            target = target + '/rrdTemplates'

        template_data = self._router_request(
            self._make_request_data(
                'addTemplate',
                dict(
                    id=name,
                    targetUid=target,
                )
            )
        )

        return self._get_template_by_uid(template_data['nodeConfig']['uid'].replace('/zport/dmd/', '', 1))

    def delete_template(self, device_class, template):
        """
        Removes a template.

        Arguments:
            device_class (str): Name of the device class where the template is defined
            template (str): Name of the template to remove

        Returns:
             dict:
        """
        if not device_class.startswith('Devices'):
            if device_class.startswith('/'):
                device_class = 'Devices{0}'.format(device_class)
            else:
                device_class = 'Devices/{0}'.format(device_class)

        template_uid = '{0}/rrdTemplates/{1}'.format(device_class, template)
        return self._router_request(
            self._make_request_data(
                'deleteTemplate',
                dict(uid=template_uid),
            )
        )

    def add_local_template(self, zenoss_object, name):
        """
        Adds a local template to an object.

        Arguments:
             zenoss_object (str): Uid of the object to add the local template to
             name (str): Unique name for the new local template

         Returns:
             ZenossTemplate:
        """
        if not zenoss_object.startswith('Devices'):
            if zenoss_object.startswith('/'):
                zenoss_object = 'Devices{0}'.format(zenoss_object)
            else:
                zenoss_object = 'Devices/{0}'.format(zenoss_object)

        template_data = self._router_request(
            self._make_request_data(
                'addLocalRRDTemplate',
                dict(
                    uid=zenoss_object,
                    templateName=name,
                )
            )
        )

        return self._get_template_by_uid(template_data['nodeConfig']['uid'].replace('/zport/dmd/', '', 1))

    def delete_local_template(self, zenoss_object, name):
        """
        Builds the request data for deleting a local template to an object.

        Arguments:
             object (str): Uid of the object to remove the local template from
             name (str): Unique name of the new local template

         Returns:
             dict:
        """
        if not zenoss_object.startswith('Devices'):
            if zenoss_object.startswith('/'):
                zenoss_object = 'Devices{0}'.format(zenoss_object)
            else:
                zenoss_object = 'Devices/{0}'.format(zenoss_object)

        return self._router_request(
            self._make_request_data(
                'removeLocalRRDTemplate',
                dict(
                    uid=zenoss_object,
                    templateName=name,
                )
            )
        )

    def get_data_source_types(self):
        """
        Gets the list of available data source types.

        Returns:
            list:
        """
        ds_types_data = self._router_request(
            self._make_request_data(
                'getDataSourceTypes',
                dict(query=''),
            )
        )
        data_source_types = []
        for ds_type in ds_types_data['data']:
            data_source_types.append(ds_type['type'])

        return data_source_types

    def get_threshold_types(self):
        """
        Gets the list of available threshold types.

        Returns:
            list:
        """
        threshold_types_data = self._router_request(
            self._make_request_data(
                'getThresholdTypes',
                dict()
            )
        )
        threshold_types = []
        for threshold_type in threshold_types_data['data']:
            threshold_types.append(threshold_type['type'])

        return threshold_types

    def add_data_point_to_graph(self, datapoint, graph, include_thresholds=False):
        """
        Adds a data point to a graph.

        Arguments:
            datapoint (str): Uid of the data point to add
            graph (str): Uid of the graph to add the data point to
            include_thresholds (bool): Set to True to include the related
                                       thresholds for the data point

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'addDataPointToGraph',
                dict(
                    dataPointUid=datapoint,
                    graphUid=graph,
                    includeThresholds=include_thresholds,
                )
            )
        )


class ZenossTemplate(TemplateRouter):
    """
    Class for Zenoss Template objects
    """

    def __init__(self, url, headers, ssl_verify, template_data):
        super(ZenossTemplate, self).__init__(url, headers, ssl_verify)

        self.definition = template_data.setdefault('definition', None)
        self.description = template_data.setdefault('description', None)
        self.hidden = template_data.setdefault('hidden', False)
        self.iconCls = template_data.setdefault('iconCls', None)
        self.id = template_data.setdefault('id', None)
        self.inspector_type = template_data.setdefault('inspector_type', None)
        self.leaf = template_data.setdefault('leaf', True)
        self.meta_type = template_data.setdefault('meta_type', 'RRDTemplate')
        self.name = template_data['name']
        self.qtip = template_data.setdefault('qtip', None)
        self.targetPythonClass = template_data.setdefault('targetPythonClass', None)
        self.text = template_data.setdefault('text', None)

        if 'uid' in template_data:
            self.uid = template_data['uid'].replace('/zport/dmd/', '', 1)
        else:
            self.uid = None

    def copy(self, target):
        """
        Copy a template to another device or device class.

        Arguments:
            target (str): Uid of the device or device class to copy to

        Returns:
            ZenossTemplate:
        """
        if not target.endswith('rrdTemplates'):
            target = target + '/rrdTemplates'
        template_data = self._router_request(
            self._make_request_data(
                'copyTemplate',
                dict(
                    uid=self.uid,
                    targetUid=target,
                )
            )
        )
        return ZenossTemplate(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            template_data['data']
        )

    def delete(self):
        """
        Removes a template.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteTemplate',
                dict(uid=self.uid),
            )
        )

    def get_data_sources(self):
        """
        Gets data sources configured for a template.

        Returns:
            list(ZenossDataSource):
        """
        ds_data = self._router_request(
            self._make_request_data(
                'getDataSources',
                dict(uid=self.uid),
            )
        )
        datasources = []
        for ds in ds_data['data']:
            datasources.append(
                ZenossDataSource(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    ds
                )
            )
        return datasources

    def list_data_sources(self):
        """
        Rerturns data sources configured for a template as a list.

        Returns:
            list(str):
        """
        ds_data = self._router_request(
            self._make_request_data(
                'getDataSources',
                dict(uid=self.uid),
            )
        )
        datasources = []
        for ds in ds_data['data']:
            datasources.append(ds['uid'].replace('/zport/dmd/', '', 1))

        return datasources

    def get_data_source(self, datasource):
        """
        Get a particular data source.

        Arguments:
            datasource (str): Name of the data source to get

        Returns:
            ZenossDataSource:
        """
        datasource_uid = '{}/datasources/{}'.format(self.uid, datasource)

        return self._get_data_source_by_uid(datasource_uid)

    def add_data_source(self, datasource, type):
        """
        Adds a data source to a template.

        Arguments:
            datasource (str): Name of the new data source
            type (str): Type of the new data source, must be one of the types
                        returned by get_data_source_types()

        Returns:
            ZenossDataSource:
        """
        response_data = self._router_request(
            self._make_request_data(
                'addDataSource',
                dict(
                    templateUid=self.uid,
                    name=datasource,
                    type=type,
                )
            )
        )
        return self._get_data_source_by_uid(
            '{}/datasources/{}'.format(self.uid, datasource)
        )

    def delete_data_source(self, datasource):
        """
        Deletes a data source from a template.

        Arguments:
            datasource (str): Name the data source to remove

        Returns:
            dict:
        """
        datasource_uid = '{}/datasources/{}'.format(self.uid, datasource)
        return self._router_request(
            self._make_request_data(
                'deleteDataSource',
                dict(uid=datasource_uid),
            )
        )

    def get_data_points(self):
        """
        Get all the data points in a template.

        Returns:
            list(ZenossDataPoint):
        """
        dp_data = self._router_request(
            self._make_request_data(
                'getDataPoints',
                dict(uid=self.uid)
            )
        )

        datapoints = []
        for dp in dp_data['data']:
            datapoints.append(
                ZenossDataPoint(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    dp
                )
            )
        return datapoints

    def list_data_points(self):
        """
        Returns all the data points in a template as a list.

        Returns:
            list(str):
        """
        dp_data = self._router_request(
            self._make_request_data(
                'getDataPoints',
                dict(uid=self.uid)
            )
        )

        datapoints = []
        for dp in dp_data['data']:
            datapoints.append(dp['uid'].replace('/zport/dmd/', '', 1))

        return datapoints

    def get_thresholds(self):
        """
        Gets the thresholds of a template.

        Returns:
            list(ZenossThresholds):
        """
        threshold_data = self._router_request(
            self._make_request_data(
                'getThresholds',
                dict(uid=self.uid),
            )
        )

        thresholds = []
        for t in threshold_data['data']:
            thresholds.append(
                ZenossThreshold(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    t
                )
            )

        return thresholds

    def list_thresholds(self):
        """
        Returns the thresholds of a template as a list.

        Returns:
            list(str):
        """
        threshold_data = self._router_request(
            self._make_request_data(
                'getThresholds',
                dict(uid=self.uid),
            )
        )

        thresholds = []
        for t in threshold_data['data']:
            thresholds.append(t['uid'].replace('/zport/dmd/', '', 1))

        return thresholds

    def get_threshold(self, threshold):
        """
        Get a particular threshold.

        Arguments:
            threshold (str): Name of the threshold to get details on

        Returns:
            ZenossThreshold:
        """
        threshold_uid = '{}/thresholds/{}'.format(self.uid, threshold)
        return self._get_threshold_by_uid(threshold_uid)

    def add_threshold(self, threshold, threshold_type, datapoints):
        """
        Adds a threshold to a template.

        Arguments:
            threshold (str): Name of the new threshold
            threshold_type (str): Type of the new threshold, must be one of the types
                        returned by get_threshold_types()
            datapoints (list): List of datapoints to select for the threshold

        Returns:
            ZenossThreshold:
        """
        if not isinstance(datapoints, list):
            raise ZenossAPIClientError('Type error: datapoints to add to a threshold must be a list')

        response_data = self._router_request(
            self._make_request_data(
                'addThreshold',
                dict(
                    uid=self.uid,
                    thresholdId=threshold,
                    thresholdType=threshold_type,
                    dataPoints=datapoints,
                )
            )
        )

        return self._get_threshold_by_uid(
            '{}/thresholds/{}'.format(self.uid, threshold)
        )

    def delete_threshold(self, threshold):
        """
        Deletes a threshold.

        Arguments:
            threshold (str): Name of the threshold to remove

        Returns:
            dict:
        """
        threshold_uid = '{}/thresholds/{}'.format(self.uid, threshold)
        return self._router_request(
            self._make_request_data(
                'removeThreshold',
                dict(uid=threshold_uid),
            )
        )

    def get_graphs(self):
        """
        Get the graphs defined for a template.

        Returns:
            list(ZenossGraph):
        """
        graphs_data = self._router_request(
            self._make_request_data(
                'getGraphs',
                dict(uid=self.uid),
            )
        )

        graphs = []
        for g in graphs_data:
            graphs.append(
                ZenossGraph(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    g
                )
            )
        return graphs

    def list_graphs(self):
        """
        Returns the graphs defined for a template as a list.

        Returns:
            list(str):
        """
        graphs_data = self._router_request(
            self._make_request_data(
                'getGraphs',
                dict(uid=self.uid),
            )
        )

        graphs = []
        for g in graphs_data:
            graphs.append(g['uid'].replace('/zport/dmd/', '', 1))

        return graphs

    def get_graph(self, graph):
        """
        Get a particular graph.

        Arguments:
            graph (str): Name of the graph to get the definition of

        Returns:
            ZenossGraph:
        """
        graph_uid = '{}/graphDefs/{}'.format(self.uid, graph)

        return self._get_graph_by_uid(graph_uid)

    def add_graph(self, graph):
        """
        Add a new graph to a template.

        Arguments:
            graph (str): Name for the new graph

        Returns:
            ZenossGraph:
        """
        response_data = self._router_request(
            self._make_request_data(
                'addGraphDefinition',
                dict(
                    templateUid=self.uid,
                    graphDefinitionId=graph,
                )
            )
        )

        return self._get_graph_by_uid(
            '{}/graphDefs/{}'.format(self.uid, graph)
        )

    def delete_graph(self, graph):
        """
        Delete a particular graph.

        Arguments:
            graph (str): The name of the graph to delete.

        Returns:
            dict:
        """
        graph_uid = '{0}/graphDefs/{1}'.format(self.uid, graph)

        return self._router_request(
            self._make_request_data(
                'deleteGraphDefinition',
                data=dict(
                    uid=graph_uid,
                )
            )
        )


class ZenossDataSource(TemplateRouter):
    """
    Class for Zenoss template data sources
    """

    def __init__(self, url, headers, ssl_verify, ds_data):
        super(ZenossDataSource, self).__init__(url, headers, ssl_verify)

        uid = ds_data.pop('uid')
        self.uid = uid.replace('/zport/dmd/', '', 1)
        self.name = ds_data.pop('name')
        self.id = ds_data.pop('id')
        self.eventClass = ds_data.pop('eventClass')
        self.eventKey = ds_data.pop('eventKey')
        self.severity = ds_data.pop('severity')
        self.type = ds_data.pop('type')
        self.component = ds_data.pop('component')
        self.properties = ds_data
        self.parent = self.uid.split('/datasources/')[0]

    def delete(self):
        """
        Deletes a data source from a template.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteDataSource',
                dict(uid=self.uid),
            )
        )

    def get_data_points(self):
        """
        Get all the data points for a datasource.

        Returns:
            list(ZenossDataPoint):
        """
        dp_data = self._router_request(
            self._make_request_data(
                'getDataPoints',
                dict(uid=self.parent)
            )
        )

        datapoints = []
        for dp in dp_data['data']:
            if dp['name'].startswith(self.name):
                datapoints.append(
                    ZenossDataPoint(
                        self.api_url,
                        self.api_headers,
                        self.ssl_verify,
                        dp
                    )
                )
        return datapoints

    def list_data_points(self):
        """
        Returns all the data points for a datasource as a list.

        Returns:
            list(str):
        """
        dp_data = self._router_request(
            self._make_request_data(
                'getDataPoints',
                dict(uid=self.parent)
            )
        )

        datapoints = []
        for dp in dp_data['data']:
            if dp['name'].startswith(self.name):
                datapoints.append(dp['uid'].replace('/zport/dmd/', '', 1))

        return datapoints

    def get_data_point(self, datapoint):
        """
        Get a particular data point.

        Arguments:
            datapoint (str): Name of the data point to get details for

        Returns:
            ZenossDataPoint:
        """
        datapoint_uid = '{}/datapoints/{}'.format(self.uid, datapoint)

        return self._get_data_point_by_uid(datapoint_uid)

    def add_data_point(self, datapoint):
        """
        Adds a data point to a data source.

        Arguments:
            datapoint (str): Name of the new data point

        Returns:
            ZenossDataPoint:
        """
        response_data = self._router_request(
            self._make_request_data(
                'addDataPoint',
                dict(
                    dataSourceUid=self.uid,
                    name=datapoint,
                )
            )
        )

        return self._get_data_point_by_uid(
            '{}/datapoints/{}'.format(self.uid, datapoint)
        )

    def delete_data_point(self, datapoint):
        """
        Deletes a data point from a template.

        Arguments:
            datapoint (str): Name of the data point to remove

        Returns:
            dict:
        """
        datapoint_uid = '{}/datapoints/{}'.format(self.uid, datapoint)
        return self._router_request(
            self._make_request_data(
                'deleteDataPoint',
                dict(uid=datapoint_uid),
            )
        )


class ZenossDataPoint(TemplateRouter):
    """
    Class for Zenoss data points
    """

    def __init__(self, url, headers, ssl_verify, dp_data):
        super(ZenossDataPoint, self).__init__(url, headers, ssl_verify)

        self.isrow = dp_data['isrow']
        self.leaf = dp_data['leaf']
        self.description = dp_data['description']
        self.rrdmin = dp_data['rrdmin']
        self.name = dp_data['name']
        self.rate = dp_data['rate']
        self.newId = dp_data['newId']
        self.createCmd = dp_data['createCmd']
        self.rrdtype = dp_data['rrdtype']
        self.rrdmax = dp_data['rrdmax']
        self.aliases = dp_data['aliases']
        self.type = dp_data['type']
        self.id = dp_data['id']
        self.uid = dp_data['uid'].replace('/zport/dmd/', '', 1)
        self.parent = self.uid.split('/datapoints/')[0]

    def set_threshold(self, threshold, threshold_type):
        """
        Adds a threshold for the data point

        Arguments:
            threshold (str): Name of the threshold to add
            threshold_type (str): Type of the new threshold, must be one of the
                        types returned by get_threshold_types()

        Returns:
            ZenossThreshold:
        """
        template = self.uid.split('/datasources/')[0]

        response_data = self._router_request(
            self._make_request_data(
                'addThreshold',
                dict(
                    uid=template,
                    thresholdId=threshold,
                    thresholdType=threshold_type,
                    dataPoints=[self.uid],
                )
            )
        )

        return self._get_threshold_by_uid(
            '{}/thresholds/{}'.format(self.uid, threshold)
        )

    def make_counter(self):
        """
        Sets the RRD Type of the data point to COUNTER

        Returns:
            bool:
        """
        self.set_properties(dict(rrdtype='COUNTER'))
        self.rrdtype = 'COUNTER'

        return True

    def make_gauge(self):
        """
        Sets the RRD Type of the data point to GAUGE

        Returns:
            bool:
        """
        self.set_properties(dict(rrdtype='GAUGE'))
        self.rrdtype = 'GAUGE'

        return True

    def add_to_graph(self, graph, include_thresholds=False):
        """
        Adds a data point to a graph.

        Arguments:
            graph (str): Name of the graph to add the data point to
            include_thresholds (bool): Set to True to include the related
                                       thresholds for the data point

        Returns:
            dict:
        """
        graph_uid = '{0}/graphDefs/{1}'.format(self.uid.split('/datasources/')[0], graph)
        return self.add_data_point_to_graph(self.uid, graph_uid, include_thresholds)

    def delete(self):
        """
        Deletes a data point from a template.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteDataPoint',
                dict(uid=self.uid),
            )
        )


class ZenossThreshold(TemplateRouter):
    """
    Class for Zenoss thresholds
    """

    def __init__(self, url, headers, ssl_verify, threshold_data):
        super(ZenossThreshold, self).__init__(url, headers, ssl_verify)

        self.description = threshold_data.pop('description')
        self.enabled = threshold_data.pop('enabled')
        self.name = threshold_data.pop('name')
        self.dataPoints = threshold_data.pop('dataPoints')
        self.eventClass = threshold_data.pop('eventClass')
        self.type = threshold_data.pop('type')
        self.id = threshold_data.pop('id')
        self.severity = threshold_data.pop('severity')
        uid = threshold_data.pop('uid')
        self.uid = uid.replace('/zport/dmd/', '', 1)
        self.parent = self.uid.split('/thresholds/')[0]
        self.properties = threshold_data

    def set_max(self, maxval):
        """
        Sets the threshold value for a MinMaxThreshold checking the max
        value of a data point.

        Arguments:
            maxval (str): Maximum value for the data point before alerting

        Returns:
            bool:
        """
        self.set_properties(dict(maxval=maxval))
        self.properties['maxval'] = maxval

        return True

    def set_min(self, minval):
        """
        Sets the threshold value for a MinMaxThreshold checking the minimum
        value of a data point.

        Arguments:
            minval (str): Minimum value for the data point before alerting

        Returns:
            bool:
        """
        self.set_properties(dict(minval=minval))
        self.properties['minval'] = minval

        return True

    def delete(self):
        """
        Deletes a threshold.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'removeThreshold',
                dict(uid=self.uid),
            )
        )


class ZenossGraph(TemplateRouter):
    """
    Class for Zenoss graphs
    """

    def __init__(self, url, headers, ssl_verify, graph_data):
        super(ZenossGraph, self).__init__(url, headers, ssl_verify)

        self.sequence = graph_data['sequence']
        self.height = graph_data['height']
        self.miny = graph_data['miny']
        self.id = graph_data['id']
        self.maxy = graph_data['maxy']
        self.autoscale = graph_data['autoscale']
        self.log = graph_data['log']
        self.custom = graph_data['custom']
        self.width = graph_data['width']
        self.graphPoints = graph_data['graphPoints']
        self.rrdVariables = graph_data['rrdVariables']
        self.units = graph_data['units']
        self.hasSummary = graph_data['hasSummary']
        self.ceiling = graph_data['ceiling']
        self.description = graph_data['description']
        self.base = graph_data['base']
        self.name = graph_data['name']
        self.uid = graph_data['uid'].replace('/zport/dmd/', '', 1)
        self.parent = self.uid.split('/graphDefs/')[0]

    def delete(self):
        """
        Delete the graph.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteGraphDefinition',
                data=dict(
                    uid=self.uid,
                )
            )
        )

    def get_points(self):
        """
        Gets the data points of a graph.

        Returns:
            list(ZenossDataPoint):
        """
        points_data = self._router_request(
            self._make_request_data(
                'getGraphPoints',
                dict(uid=self.uid,)
            )
        )

        points = []
        for p in points_data['data']:
            point_datasource = p['dpName'].split('_')[0]
            points.append(
                self._get_data_point_by_uid('{0}/datasources/{1}/datapoints/{2}'.format(self.parent, point_datasource, p['dpName']))
            )
        return points

    def list_points(self):
        """
        Returns the data points of a graph as a list.

        Returns:
            list(str):
        """
        points_data = self._router_request(
            self._make_request_data(
                'getGraphPoints',
                dict(uid=self.uid,)
            )
        )

        points = []
        for p in points_data['data']:
            points.append(p['uid'].replace('/zport/dmd/', '', 1))

        return points

    def add_point(self, datasource, datapoint, include_thresholds=False):
        """
        Adds a data point to a graph.

        Arguments:
            datasource (str): Name of the data source holding the data point
            datapoint (str): Name of the data point to add
            include_thresholds (bool): Set to True to include the related
                                       thresholds for the data point

        Returns:
            dict:
        """
        datapoint_uid = '{0}/datasources/{1}/datapoints/{2}'.format(self.parent, datasource, datapoint)
        return self.add_data_point_to_graph(datapoint_uid, self.uid, include_thresholds)

    def set_graph_properties(self, properties):
        """
        Set the properties for a graph.

        Arguments:
            properties (dict): Properties and values to set

        Returns:
            dict:
        """
        data = dict(uid=self.uid)
        data.update(properties)
        response_data = self._router_request(
            self._make_request_data(
                'setGraphDefinition',
                data,
            )
        )
        for prop in properties:
            if getattr(self, prop, False):
                setattr(self, prop, properties[prop])

        return response_data

    def set_zero_baseline(self):
        """
        Set the minimum value of a graph display to zero. By default
        Zenoss graph scale is dynamic, meaning the display
        can be skewed because the minimum value isn't fixed.
        """
        return self.set_graph_properties(
            properties=dict(miny=0),
        )

    def set_point_sequence(self, datapoints):
        """
        Sets the order of data points in a graph.

        Arguments:
            datapoints (list): List of data point names in the desired order

        Returns:
            dict:
        """
        if not isinstance(datapoints, list):
            raise ZenossAPIClientError('Type error: Sequence of datapoints must be a list')

        graph_points = []
        for p in datapoints:
            graph_points.append('{}/graphPoints/{}'.format(self.uid, p))

        return self._router_request(
            self._make_request_data(
                'setGraphPointSequence',
                dict(uids=graph_points)
            )
        )

    def delete_point(self, datapoint):
        """
        Deletes a data point from a graph.

        Arguments:
            datapoint (str): Name of the data point to remove

        Returns:
            dict:
        """
        response_data = self._router_request(
            self._make_request_data(
                'deleteGraphPoint',
                dict(uid='{}/graphPoints/{}'.format(self.uid, datapoint))
            )
        )

        graph_data = self._router_request(
            self._make_request_data(
                'getGraphDefinition',
                dict(uid=self.uid)
            )
        )

        self.graphPoints = graph_data['data']['graphPoints']
        self.rrdVariables = graph_data['data']['rrdVariables']

        return response_data

    def add_graph_threshold(self, threshold):
        """
        Adds a threshold to a graph.

        Arguments:
            threshold (str): Uid of the threshold to add

        Returns:
            dict:
        """
        response_data = self._router_request(
            self._make_request_data(
                'addThresholdToGraph',
                dict(
                    graphUid=self.uid,
                    thresholdUid=threshold,
                )
            )
        )

        graph_data = self._router_request(
            self._make_request_data(
                'getGraphDefinition',
                dict(uid=self.uid)
            )
        )

        self.graphPoints = graph_data['data']['graphPoints']

        return response_data



from setuptools import find_packages, setup

name = 'ZenossAPIClient'
version = '0.2'
release = '0.2.2'

setup(name=name,
      version=release,
      description='Zenoss API client module',
      long_description="""
Python module for interacting with the Zenoss API an an object-oriented way.
The philosophy here is to use objects to work with everything in the Zenoss 
API, and to try to normalize the various calls to the different routers.
Thus `get` methods will always return an object, `list` methods will return 
data. All methods to add or create start with `add`, all remove or delete 
start with `delete`. As much as possible the methods try to hide the 
idiosyncrasies of the JSON API, and to do the work for you, for example 
by letting you use a device name instead of having to provide the full 
device UID for every call.
""",
      url='https://github.com/Zuora-TechOps/ZenossAPIClient',
      author='Mark Troyer',
      author_email='disco@zuora.com',
      license='Apache',
      classifiers=[
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Apache Software License',
          'Natural Language :: English',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Topic :: System :: Monitoring',
      ],
      keywords='Zenoss monitoring api',
      packages=find_packages(where='src'),
      package_dir={'': 'src'},
      zip_safe=False,
      tests_require=[
          'pytest>=3.2.3',
          'pytest-responses>=0.3.0',
      ],
      install_requires=[
          'python-dateutil>=2.6.1',
          'requests>=2.18.1',
      ]
      )



# -*- coding: utf-8 -*-

"""
Zenoss devicemanagement_router
"""

from datetime import datetime as dt
from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter


class DeviceManagementRouter(ZenossRouter):
    """
    Class for interacting with the Zenoss devicemanagement router
    """

    def __init__(self, url, headers, ssl_verify):
        super(DeviceManagementRouter, self).__init__(url, headers, ssl_verify,
                                                     'devicemanagement_router',
                                                     'DeviceManagementRouter')
        self.uid = None

    def __repr__(self):
        if self.uid:
            identifier = self.uid
        else:
            identifier = hex(id(self))

        return '<{0} object at {1}>'.format(
            type(self).__name__, identifier
        )

    def timezone(self):
        """
        Returns the configured timezone.

        Returns:
            str:
        """
        tz_data = self._router_request(
            self._make_request_data(
                'getTimeZone',
                data=dict()
            )
        )

        return tz_data['data']

    def list_maintenance_windows(self, uid):
        """
        Returns the list of maintenance windows configured for a device
        or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(dict):
        """
        uid = self._check_uid(uid)

        mw_data = self._router_request(
            self._make_request_data(
                'getMaintWindows',
                data=dict(
                    uid=uid
                )
            )
        )

        for mw in mw_data['data']:
            mw['uid'] = mw['uid'].replace('/zport/dmd/', '', 1)

        return mw_data['data']

    def get_maintenance_windows(self, uid):
        """
        Returns a list of ZenossMaintenanceWindow objects for the
        maintenance windows configured for a device or device class

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(ZenossMaintenanceWindow):
        """
        mw_data = self.list_maintenance_windows(uid)
        windows = []
        for mw in mw_data:
            windows.append(
                ZenossMaintenanceWindow(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    mw,
                    parent=self._check_uid(uid),
                )
            )

        return windows

    def get_maintenance_window(self, uid, name):
        """
        Get a maintenance window object for the named window.

        Arguments:
            uid (str): The UID of the device or device class
            name (str): Name of the maintenance window

        Returns:
            ZenossMaintenanceWindow:
        """
        mw_data = self.list_maintenance_windows(uid)

        for mw in mw_data:
            if mw['name'] == name:
                return ZenossMaintenanceWindow(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    mw,
                    parent=self._check_uid(uid),
                )

        return None

    def add_maintenance_window(self, uid, name, start, duration, enabled=False, start_state=300, repeat='Never', occurrence='1st', days='Sunday'):
        """
        Add a new maintenance window for device or device class.

        Arguments:
            uid (str): The UID of the device or device class
            start (str): Window start time in UNIX epoch timestamp format,
                e.g. "1511290393"
            duration (str): Duration of the window in HH:MM:SS format
            start_state (int): Production state for the maintenance window,
                default is 300 (Maintenance)
            repeat (str): Maintenance window repeat interval, default is
                'Never'. Other valid choices are: 'Daily', 'Every Weekday',
                'Weekly', 'Monthly: day of month', 'Monthly: day of week'
            occurrence (str): For 'Monthly: day of week' repeats, options are
                '1st', '2nd', '3rd', '4th', '5th', 'Last'
            days (str): For 'Monthly: day of week' repeats, options are
                'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
                'Saturday', 'Sunday'

        Returns:
            ZenossMaintenanceWindow:
        """
        if repeat not in ['Never', 'Daily', 'Every Weekday', 'Weekly',
                          'Monthly: day of month', 'Monthly: day of week']:
            raise ZenossAPIClientError(
                'Invalid maintenance window repetition: {0}'.format(repeat))

        if occurrence not in ['1st', '2nd', '3rd', '4th', '5th', 'Last']:
            raise ZenossAPIClientError(
                'Invalid maintenance window occurrence: {0}'.format(occurrence))

        if days not in ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
                        'Saturday', 'Sunday']:
            raise ZenossAPIClientError(
                'Invalid maintenance window days: {0}'.format(days))

        self._router_request(
            self._make_request_data(
                'addMaintWindow',
                data=dict(
                    params=dict(
                        uid=uid,
                        name=name,
                        start=start,
                        duration=duration,
                        enabled=enabled,
                        startState=start_state,
                        repeat=repeat,
                        occurrence=occurrence,
                        days=days,
                    )
                )
            )
        )

        return self.get_maintenance_window(uid, name)

    def list_user_commands(self, uid):
        """
        Get the list of user commands configured for a device or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(dict):
        """
        uid = self._check_uid(uid)

        uc_data = self._router_request(
            self._make_request_data(
                'getUserCommands',
                data=dict(
                    uid=uid,
                )
            )
        )

        return uc_data['data']

    def get_user_commands(self, uid):
        """
        Get a list of user commands objects configured for a device
        or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(ZenossUserCommand):
        """
        uc_data = self.list_user_commands(uid)
        user_commands = []
        for uc in uc_data:
            user_commands.append(
                ZenossUserCommand(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    uc,
                    parent=self._check_uid(uid)
                )
            )

        return user_commands

    def get_user_command_by_id(self, uid, command_id):
        """
        Get a configured user command by its id

        Arguments:
            uid (str): The UID of the device or device class
            command_id (str): The ID of the user command
        """
        uc_data = self.list_user_commands(uid)
        for uc in uc_data:
            if uc['id'] == command_id:
                return ZenossUserCommand(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    uc,
                    parent=self._check_uid(uid)
                )

        return None

    def get_user_command_by_name(self, uid, command_name):
        """
        Get a configured user command by its id

        Arguments:
            uid (str): The UID of the device or device class
            command_name (str): The name of the user command
        """
        uc_data = self.list_user_commands(uid)
        for uc in uc_data:
            if uc['name'] == command_name:
                return ZenossUserCommand(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    uc,
                    parent=self._check_uid(uid)
                )

        return None

    def add_user_command(self, uid, name, description, command, password):
        """
        Add a new user command to a device or device class.

        Arguments:
            uid (str): The UID of the device or device class
            name (str): Name for the new command
            description (str): Description of the new command
            command (str): Command line of the new command, can include TALES
                expressions
            password (str): Password of the user adding the command.
        """
        uid = self._check_uid(uid)
        self._router_request(
            self._make_request_data(
                'addUserCommand',
                data=dict(
                    uid=uid,
                    name=name,
                    description=description,
                    command=command,
                    password=password
                )
            )
        )

        return self.get_user_command_by_name(uid, name)

    def list_users(self, uid):
        """
        List the users available to associate with a device or device class.

        Arguments:
            uid (str): the UID of the device or device class

        Returns:
            list(str):
        """
        uid = self._check_uid(uid)
        user_data = self._router_request(
            self._make_request_data(
                'getUserList',
                data=dict(
                    uid=uid
                )
            )
        )
        return user_data['data']

    def list_available_roles(self, uid):
        """
        List the admin roles available to associate with a device or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(str):
        """
        uid = self._check_uid(uid)
        role_data = self._router_request(
            self._make_request_data(
                'getRolesList',
                data=dict(
                    uid=uid
                )
            )
        )

        return role_data['data']

    def list_admin_roles(self, uid):
        """
        List the admin roles associated with a device or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(dict):
        """
        uid = self._check_uid(uid)
        role_data = self._router_request(
            self._make_request_data(
                'getAdminRoles',
                data=dict(
                    uid=uid
                )
            )
        )

        for r in role_data['data']:
            r['uid'] = r['uid'].replace('/zport/dmd/', '', 1)

        return role_data['data']

    def get_admins(self, uid):
        """
        Get ZenossDeviceManagementAdmin objects for the configured admin
        users for a device or device class.

        Arguments:
            uid (str): The UID of the device or device class

        Returns:
            list(ZenossDeviceManagementAdmin):
        """
        admin_data = self.list_admin_roles(uid)
        admins = []
        for admin in admin_data:
            admins.append(
                ZenossDeviceManagementAdmin(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    admin
                )
            )

        return admins

    def list_admins_by_role(self, uid, role):
        """
        List configured admin users for a device or device class by role.

        Arguments:
            uid (str): The UID of the device or device class
            role (str): The role to filter on

        Returns:
            list(dict):
        """
        admin_data = self.list_admin_roles(uid)
        admins = []
        for admin in admin_data:
            if admin['role'] == role:
                admins.append(admin)

        return admins

    def get_admins_by_role(self, uid, role):
        """
        Get ZenossDeviceManagementAdmin objects for the configured admin users
        of a device or device class by role.

        Arguments:
            uid (str): The UID of the device or device class
            role (str): The role to filter on

        Returns:
            list(ZenossDeviceManagementAdmin):
        """
        admin_data = self.list_admin_roles(uid)
        admins = []
        for admin in admin_data:
            if admin['role'] == role:
                admins.append(
                    ZenossDeviceManagementAdmin(
                        self.api_url,
                        self.api_headers,
                        self.ssl_verify,
                        admin
                    )
                )

        return admins

    def get_admin_by_name(self, uid, name):
        """
        Get an admin user for a device or device class by name.

        Arguments:
            uid (str): The UID of the device or device class
            name (str): The name of the admin user

        Returns:
            ZenossDeviceManagementAdmin:
        """
        admin_data = self.list_admin_roles(uid)
        for admin in admin_data:
            if admin['name'] == name:
                return ZenossDeviceManagementAdmin(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    admin
                )

        return None

    def get_admin_by_id(self, uid, admin_id):
        """
        Get and admin user for a device or device class by id.

        Arguments:
            uid (str): The UID of the device or device class
            admin_id (str): The ID of the admin user

        Returns:
            ZenossDeviceManagementAdmin:
        """
        admin_data = self.list_admin_roles(uid)
        for admin in admin_data:
            if admin['id'] == admin_id:
                return ZenossDeviceManagementAdmin(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    admin
                )

        return None

    def add_admin(self, uid, name, role=None):
        """
        Add an admin user to a device or device class.

        Arguments:
            uid (str): The UID of the device or device class
            name (str): The name of the user to add
            role (str): The role to associate with the user for this
                device or device class

        Returns:
            ZenossDeviceManagementAdmin:
        """
        uid = self._check_uid(uid)
        self._router_request(
            self._make_request_data(
                'addAdminRole',
                data=dict(
                    params=dict(
                        uid=uid,
                        name=name,
                        role=role,
                    )
                )
            )
        )

        return self.get_admin_by_name(uid, name)


class ZenossMaintenanceWindow(DeviceManagementRouter):
    """
    Class for Zenoss maintenance window objects
    """

    def __init__(self, url, headers, ssl_verify, window_data, parent=None):
        super(ZenossMaintenanceWindow, self).__init__(url, headers, ssl_verify)

        self.uid = window_data['uid'].replace('/zport/dmd/', '', 1)
        self.id = window_data['id']
        self.name = window_data['name']
        self.description = window_data['description']
        self.meta_type = window_data['meta_type']
        self.inspector_type = window_data['inspector_type']
        self.startState = window_data['startState']
        self.startProdState = window_data['startProdState']
        self.enabled = window_data['enabled']
        self.started = window_data['started']
        self.start = window_data['start']
        self.startTime = window_data['startTime']
        self.duration = window_data['duration']
        self.skip = window_data['skip']
        self.repeat = window_data['repeat']
        self.niceRepeat = window_data['niceRepeat']
        self.occurrence = window_data['occurrence']
        self.days = window_data['days']
        self.parent = parent

        self.startDate, start_ts, tz = self.startTime.split()
        self.startHours, self.startMinutes = start_ts.split(':')[:2]
        duration_parts = self.duration.split()
        if len(duration_parts) > 1:
            self.durationDays = duration_parts[0]
        else:
            self.durationDays = None
        self.durationHours, self.durationMinutes = duration_parts[-1].split(':')[:2]

    def _build_update_params(self, params):
        """
        Build the params dict for updating a maintenance window

        Arguments:
            params (dict): The update parameters as a dict

        Returns:
            dict:
        """

    def delete(self):
        """
        Delete a maintenance window from a device or device class.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteMaintWindow',
                data=dict(
                    uid=self.parent,
                    id=self.id,
                )
            )
        )

    def update(self, start_timestamp=None, start_datetime=None, start_date=None,
               start_hours=None, start_minutes=None, duration_days=None,
               duration_time=None, duration_hours=None, duration_minutes=None,
               production_state=None, enabled=None, repeat=None,
               occurrence=None, days=None):
        """
        Update the settings for a maintenance window, with flexible options for
        specifying the start date/time and duration.

        Arguments:
            start_timestamp (float): Start date and time in UNIX timestamp format
            start_datetime (datetime): Start date and time as a datetime.datetime object
            start_date (str): Start date as a string
            start_hours (str): Start hours as a string
            start_minutes (str): Start minutes as a string
            duration_days (str): Duration days
            duration_time (str): Duration time in "HH:MM" format
            duration_hours (str): Duration hours
            duration_minutes (str): Duration minutes
            production_state (int): Production state for the window
            enabled (bool): Enabled state of the window
            occurrence (str): Repeat occurrence
            days (str): Repeat days

        Returns:
            bool:
        """
        new_params = dict()
        if start_timestamp:
            new_start = dt.fromtimestamp(start_timestamp)
            new_params['startDate'] = str(new_start.date())
            new_params['startHours'] = str(new_start.hour)
            new_params['startMinutes'] = str(new_start.minute)
        elif start_datetime:
            new_params['startDate'] = str(start_datetime.date())
            new_params['startHours'] = str(start_datetime.hour)
            new_params['startMinutes'] = str(start_datetime.minute)
        if start_date:
            new_params['startDate'] = start_date
        if start_hours:
            new_params['startHours'] = start_hours
        if start_minutes:
            new_params['startMinutes'] = start_minutes

        new_params['durationDays'] = duration_days
        if duration_time:
            new_params['durationHours'], new_params['durationMinutes'] = duration_time.split(':')[:2]
        if duration_hours:
            new_params['durationHours'] = duration_hours
        if duration_minutes:
            new_params['durationMinutes'] = duration_minutes

        update_resp = self._router_request(
            self._make_request_data(
                'editMaintWindow',
                data=dict(
                    params=dict(
                        uid=self.parent,
                        id=self.id,
                        startDate=new_params.get('startDate', self.startDate),
                        startHours=new_params.get('startHours', self.startHours),
                        startMinutes=new_params.get('startMinutes', self.startMinutes),
                        durationDays=new_params.get('durationDays', self.durationDays),
                        durationHours=new_params.get('durationHours', self.durationHours),
                        durationMinutes=new_params.get('durationMinutes', self.durationMinutes),
                        startProductionState=production_state if production_state else self.startProdState,
                        repeat=repeat if repeat else self.repeat,
                        enabled=enabled if enabled else self.enabled,
                        occurrence=occurrence if occurrence else self.occurrence,
                        days=days if days else self.days
                    )
                )
            )
        )

        self.__init__(self.api_url, self.api_headers, self.ssl_verify, update_resp['data'][0], parent=self.parent)

        return True

    def enable(self):
        """
        Set maintenance window to enabled.

        Returns:
            bool:
        """
        if not self.enabled:
            self._router_request(
                self._make_request_data(
                    'editMaintWindow',
                    data=dict(
                        params=dict(
                            uid=self.parent,
                            id=self.id,
                            startDate=self.startDate,
                            startHours=self.startHours,
                            startMinutes=self.startMinutes,
                            durationDays=self.durationDays,
                            durationHours=self.durationHours,
                            startProductionState=self.startProdState,
                            repeat=self.repeat,
                            enabled=True,
                            occurrence=self.occurrence,
                            days=self.days,
                        )
                    )
                )
            )
            self.enabled = True

        return True

    def disable(self):
        """
        Set maintenance window to disabled.

        Returns:
            bool:
        """
        if self.enabled:
            self._router_request(
                self._make_request_data(
                    'editMaintWindow',
                    data=dict(
                        uid=self.parent,
                        id=self.id,
                        params=dict(
                            startDate=self.startDate,
                            startHours=self.startHours,
                            startMinutes=self.startMinutes,
                            durationDays=self.durationDays,
                            durationHours=self.durationHours,
                            startProductionState=self.startProdState,
                            repeat=self.repeat,
                            enabled=False,
                            occurrence=self.occurrence,
                            days=self.days,
                        )
                    )
                )
            )
            self.enabled = False

        return True


class ZenossUserCommand(DeviceManagementRouter):
    """
    Class for Zenoss user command objects
    """

    def __init__(self, url, headers, ssl_verify, command_data, parent=None):
        super(ZenossUserCommand, self).__init__(url, headers, ssl_verify)

        self.description = command_data['description']
        self.name = command_data['name']
        self.meta_type = command_data['meta_type']
        self.command = command_data['command']
        self.inspector_type = command_data['inspector_type']
        self.id = command_data['id']
        self.uid = command_data['uid'].replace('/zport/dmd/', '', 1)
        self.parent = parent

    def delete(self):
        """
        Delete a user command for a device or device class.

        Returns:
            dict:
        """
        return self._router_request(
            self._make_request_data(
                'deleteUserCommand',
                data=dict(
                    uid=self.parent,
                    id=self.id
                )
            )
        )

    def update(self, description=None, command=None, password=None):
        """
        Update a user command.

        Arguments:
            description (str): Description of the user command
            command (str): Command line of the command
            password (str): Password of the user updating the command.
        """
        self._router_request(
            self._make_request_data(
                'updateUserCommand',
                data=dict(
                    params=dict(
                        uid=self.parent,
                        id=self.id,
                        description=description if description else self.description,
                        command=command if command else self.command,
                        password=password,
                    )
                )
            )
        )

        uc_data = self.list_user_commands(self.parent)
        for uc in uc_data:
            if uc['id'] == self.id:
                self.__init__(self.api_url, self.api_headers, self.ssl_verify, uc, parent=self.parent)

        return True


class ZenossDeviceManagementAdmin(DeviceManagementRouter):
    """
    Class for Zenoss user command objects
    """

    def __init__(self, url, headers, ssl_verify, admin_data):
        super(ZenossDeviceManagementAdmin, self).__init__(url, headers, ssl_verify)

        self.uid = admin_data['uid'].replace('/zport/dmd/', '', 1)
        self.name = admin_data['name']
        self.id = admin_data['id']
        self.meta_type = admin_data['meta_type']
        self.role = admin_data['role']
        self.inspector_type = admin_data['inspector_type']
        self.pager = admin_data['pager']
        self.email=admin_data['email']

    def update(self, role):
        """
        Update the admin user's role.

        Arguments:
            role (str): New role for the user

        Returns:
            bool:
        """
        self._router_request(
            self._make_request_data(
                'updateAdminRole',
                data=dict(
                    params=dict(
                        uid=self.uid,
                        name=self.name,
                        role=role
                    )
                )
            )
        )

        self.role = role

        return True

    def delete(self):
        """
        Delete an admin user from a device or device class.

        :return:
        """
        return self._router_request(
            self._make_request_data(
                'removeAdmin',
                data=dict(
                    uid=self.uid,
                    id=self.id,
                )
            )
        )



# -*- coding: utf-8 -*-

"""
Zenoss monitor_router
"""

from zenossapi.apiclient import ZenossAPIClientError
from zenossapi.routers import ZenossRouter


class MonitorRouter(ZenossRouter):
    """
    Class for interacting with the Zenoss monitor router
    """

    def __init__(self, url, headers, ssl_verify):
        super(MonitorRouter, self).__init__(url, headers, ssl_verify,
                                           'monitor_router', 'MonitorRouter')
        self.id = None

    def __repr__(self):
        if self.id:
            identifier = self.id
        else:
            identifier = hex(id(self))

        return '<{0} object at {1}>'.format(
            type(self).__name__, identifier
        )

    def tree(self):
        """
        Returns the full tree of hubs and collectors

        Returns:
            list(dict):
        """
        return self._router_request(
            self._make_request_data(
                'getTree',
                data=dict(id='')
            )
        )

    def list_hubs(self):
        """
        Returns the list of configured zenhubs

        Returns:
            list(dict):
        """
        hubs_data = self.tree()
        hubs = []
        for hub in hubs_data:
            hubs.append(dict(name=hub['name'], uid=hub['uid'].replace('/zport/dmd/', '', 1)))

        return hubs

    def list_collectors(self):
        """
        Returns the list of configured collectors

        Returns:
            list(dict):
        """
        hubs_data = self.tree()
        collectors = []
        for hub in hubs_data:
            for collector in hub['children']:
                collectors.append(
                    dict(
                        name=collector['name'],
                        devcount=collector['devcount'],
                        hub=hub['name'],
                        uid=collector['uid'].replace('/zport/dmd/', '', 1)
                    )
                )

        return collectors

    def get_hubs(self):
        """
        Get the configured hubs as objects

        Returns:
            list(ZenossHub):
        """
        hubs_data = self.tree()
        hubs = []
        for hub in hubs_data:
            hubs.append(
                ZenossHub(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    hub
                )
            )

        return hubs

    def get_hub(self, name):
        """
        Get a ZenossHub object

        Arguments:
            name (str): Name of the hub to get

        Returns:
            ZenossHub:
        """
        hubs_data = self.tree()
        for hub in hubs_data:
            if hub['name'] == name:
                return ZenossHub(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    hub
                )

        return None


class ZenossHub(MonitorRouter):
    """
    Class for Zenoss hub objects
    """

    def __init__(self, url, headers, ssl_verify, hub_data):
        super(ZenossHub, self).__init__(url, headers, ssl_verify)

        self.uid = hub_data['uid'].replace('/zport/dmd/', '', 1)
        self.name = hub_data['name']
        self.text = hub_data['text']
        self.id = hub_data['id']
        self.devcount = hub_data['devcount']
        self.path = hub_data['path']
        self.ccbacked = hub_data['ccbacked']
        self.collectors = []

        for c in hub_data['children']:
            self.collectors.append(c['name'])

    def get_collectors(self):
        """
        Get the hub's collectors as objects.

        Returns:
             list(ZenossCollector):
        """
        collectors = []
        for c in self.collectors:
            cinfo = self._router_request(
                self._make_request_data(
                    'getInfo',
                    data=dict(
                        id=c
                    )
                )
            )
            collectors.append(
                ZenossCollector(
                    self.api_url,
                    self.api_headers,
                    self.ssl_verify,
                    cinfo['data']
                )
            )

        return collectors

    def get_collector(self, name):
        """
        Get a ZenossCollector object

        Arguments:
            name (str): Name of the collector to get

        Returns:
            ZenossCollector:
        """
        if name in self.collectors:
            cinfo = self._router_request(
                self._make_request_data(
                    'getInfo',
                    data=dict(
                        id=name
                    )
                )
            )
            return ZenossCollector(
                self.api_url,
                self.api_headers,
                self.ssl_verify,
                cinfo['data']
            )

        return None

    def add_collector(self, name, source=None, pool=None):
        """
        Add a new collector to the hub.

        Arguments:
            name (str): Name of the new collector
            source (str): Name of the existing collector to use as a template
            pool (str): The resource pool to place the collector in

        Returns:
            ZenossCollector:
        """
        if not source:
            raise ZenossAPIClientError("You must specify a collector to use as a template.")

        if not pool:
            raise ZenossAPIClientError("You must specify a resource pool for the collector.")

        collector_params = self._router_request(
            self._make_request_data(
                'addCollector',
                data=dict(
                    id=name,
                    sourceId=source,
                    hubId=self.name,
                    poolId=pool
                )
            )
        )

        collector_data = self._router_request(
            self._make_request_data(
                'getInfo',
                data=dict(
                    id=name
                )
            )
        )

        return ZenossCollector(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
            collector_data=collector_data['data'],
            collector_params=collector_params
        )


class ZenossCollector(MonitorRouter):
    """
    Class for Zenoss collector objects
    """

    def __init__(self, url, headers, ssl_verify, collector_data, collector_params=None):
        super(ZenossCollector, self).__init__(url, headers, ssl_verify)

        self.uid = collector_data['uid'].replace('/zport/dmd/', '', 1)
        self.ccbacked = collector_data['ccbacked']
        self.name = collector_data['name']
        self.text = collector_data['text']
        self.devcount = collector_data['devcount']
        self.path = collector_data['path']
        self.id = collector_data['id']

        if not collector_params:
            collector_params = self._router_request(
                self._make_request_data(
                    'getCollector',
                    data=dict(
                        collectorString=self.name
                    )
                )
            )

        self.configCycleInterval = collector_params['data']['configCycleInterval']
        self.pingCycleInterval = collector_params['data']['pingCycleInterval']
        self.discoveryNetworks = collector_params['data']['discoveryNetworks']
        self.description = collector_params['data']['description']
        self.modelerCycleInterval = collector_params['data']['modelerCycleInterval']
        self.processCycleInterval = collector_params['data']['processCycleInterval']
        self.meta_type = collector_params['data']['meta_type']
        self.wmiqueryTimeout = collector_params['data']['wmiqueryTimeout']
        self.statusCycleInterval = collector_params['data']['statusCycleInterval']
        self.eventlogCycleInterval = collector_params['data']['eventlogCycleInterval']
        self.wmibatchSize = collector_params['data']['wmibatchSize']
        self.pingTimeOut = collector_params['data']['pingTimeOut']
        self.winCycleInterval = collector_params['data']['winCycleInterval']
        self.pingTries = collector_params['data']['pingTries']
        self.inspector_type = collector_params['data']['inspector_type']
        self.zenProcessParallelJobs = collector_params['data']['zenProcessParallelJobs']



# -*- coding: utf-8 -*-

"""
Zenoss API Client Class
"""

import inspect
import os
import urllib3


class ZenossAPIClientError(Exception):
    pass


class ZenossAPIClientAuthenticationError(Exception):
    pass


class Client(object):
    """Client class to access the Zenoss JSON API"""

    def __init__(self, host=None, user=None, password=None, ssl_verify=None):
        """
        Create the client object to communicate with Zenoss. The authentication
        and ssl parameters can be pulled from the environment so that they
        don't have to be passed in from code or command line args.

        Arguments:
            host (str): FQDN used to access the Zenoss server
            user (str): Zenoss username
            password (str): Zenoss user's password
            ssl_verify (bool): Whether to verify the SSL certificate or not.
                Set to false when using servers with self-signed certs.
        """
        if not host:
            if 'ZENOSS_HOST' in os.environ:
                host = os.environ['ZENOSS_HOST']
        if not user:
            if 'ZENOSS_USER' in os.environ:
                user = os.environ['ZENOSS_USER']
        if not password:
            if 'ZENOSS_PASSWD' in os.environ:
                password = os.environ['ZENOSS_PASSWD']

        if ssl_verify is None:
            if 'ZENOSS_SSL_VERIFY' in os.environ:
                ssl_verify = os.environ['ZENOSS_SSL_VERIFY']
            else:
                ssl_verify = True

        if isinstance(ssl_verify, str):
            if ssl_verify == "False":
                ssl_verify = False
            else:
                ssl_verify = True

        self.api_host = host
        self.api_url = 'https://{0}/zport/dmd'.format(host)
        self.api_user = user
        self.ssl_verify = ssl_verify
        self.api_headers = {"Content-Type": "application/json"}
        self.router_list = []
        self.routers = dict()

        for router in self.get_routers():
            self.router_list.append(router)
            self.routers[router] = __import__(
                'zenossapi.routers.{0}'.format(router),
                fromlist=[router])

        if self.api_user and password:
            self.api_headers.update(urllib3.make_headers(
                basic_auth='{0}:{1}'.format(self.api_user, password)))

    def get_routers(self):
        """
        Gets the list of availble Zenoss API routers

        Returns:
            list:
        """
        router_list = []
        routers_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'routers')
        file_list = os.listdir(routers_path)
        for fname in file_list:
            name, ext = fname.split('.')
            if name == "__init__":
                continue
            if ext == "py":
                router_list.append(name)

        return router_list

    def get_router(self, router):
        """
        Instantiates and returns a Zenoss router object

        Arguments:
            router (str): The API router to use
        """
        api_router = getattr(
            self.routers[router],
            '{0}Router'.format(router.capitalize()),
        )(
            self.api_url,
            self.api_headers,
            self.ssl_verify,
        )

        return api_router

    def get_router_methods(self, router):
        """
        List all available methods for an API router

        Arguments:
            router (str): The router to get methods from

        Returns:
            list:
        """
        router_methods = []
        for method in inspect.getmembers(
            getattr(self.routers[router],
                    '{0}Router'.format(router.capitalize())),
            predicate=inspect.isroutine
        ):
            if method[0].startswith('__'):
                continue
            router_methods.append(method[0])

        return router_methods
